# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import sys


def print_version(deps=False):
    import pyrockoeost
    if deps:
        print('pyrockoeost: %s' % pyrockoeost.long_version)
        try:
            import numpy
            print('numpy: %s' % numpy.__version__)
        except ImportError:
            print('numpy: N/A')

        try:
            import scipy
            print('scipy: %s' % scipy.__version__)
        except ImportError:
            print('scipy: N/A')

        try:
            import matplotlib
            print('matplotlib: %s' % matplotlib.__version__)
        except ImportError:
            print('matplotlib: N/A')

        try:
            from pyrockoeost.gui.qt_compat import Qt
            print('PyQt: %s' % Qt.PYQT_VERSION_STR)
            print('Qt: %s' % Qt.QT_VERSION_STR)
        except ImportError:
            print('PyQt: N/A')
            print('Qt: N/A')

        try:
            import vtk
            print('VTK: %s' % vtk.VTK_VERSION)
        except ImportError:
            print('VTK: N/A')

        print('python: %s.%s.%s' % sys.version_info[:3])

    elif sys.argv[1:] == ['short']:
        print(pyrockoeost.version)
    else:
        print(pyrockoeost.long_version)


if __name__ == '__main__':
    print_version(sys.argv[1:] == ['deps'])



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import logging

from ..common import ldq
from pyrockoeost.squirrel.error import ToolError

logger = logging.getLogger('psq.cli.response')

headline = 'Print instrument response information.'


def indent(prefix, nfirst, n, message):
    return '\n'.join(
        prefix + ' ' * (nfirst if i == 0 else n) + line
        for i, line in enumerate(message.splitlines()))


def make_subparser(subparsers):
    return subparsers.add_parser(
        'response',
        help=headline,
        description=headline)


def setup(parser):
    parser.add_squirrel_selection_arguments()
    parser.add_squirrel_query_arguments(without=['kinds'])

    level_choices = ['response', 'stages']

    parser.add_argument(
        '--level',
        choices=level_choices,
        dest='level',
        default='response',
        help='Set level of detail to be printed. Choices: %s'
        % ldq(level_choices))

    parser.add_argument(
        '--problems',
        action='store_true',
        dest='problems',
        default=False,
        help='Print attached warnings and error messages.')

    parser.add_argument(
        '--plot',
        action='store_true',
        dest='plot',
        default=False,
        help='Create Bode plot showing the responses.')

    parser.add_argument(
        '--fmin',
        dest='fmin',
        type=float,
        default=0.01,
        help='Minimum frequency [Hz], default: 0.01')

    parser.add_argument(
        '--fmax',
        dest='fmax',
        type=float,
        default=100.,
        help='Maximum frequency [Hz], default: 100')

    parser.add_argument(
        '--normalize',
        dest='normalize',
        action='store_true',
        help='Normalize response to be 1 on flat part.')

    parser.add_argument(
        '--save',
        dest='out_path',
        help='Save figure to file.')

    parser.add_argument(
        '--dpi',
        dest='dpi',
        type=float,
        default=100.,
        help='DPI setting for pixel image output, default: 100')

    parser.add_argument(
        '--stages',
        dest='stages',
        metavar='START:STOP',
        help='Show response of selected stages. Indexing is Python style: '
             'count starts at zero, negative values count from the end.')

    input_quantity_choices = ['displacement', 'velocity', 'acceleration']

    parser.add_argument(
        '--input-quantity',
        dest='input_quantity',
        choices=input_quantity_choices,
        metavar='QUANTITY',
        help='Show converted response for given input quantity. choices: %s'
        % ldq(input_quantity_choices))

    parser.add_argument(
        '--show-breakpoints',
        dest='show_breakpoints',
        action='store_true',
        default=False,
        help='Show breakpoints of pole-zero responses.')

    parser.add_argument(
        '--index-labels',
        dest='index_labels',
        action='store_true',
        default=False,
        help='Label graphs only by index and print details to terminal '
             'to save space when many labels would be shown. Aggregate '
             'identical responses under a common index.')


def run(parser, args):
    squirrel = args.make_squirrel()

    stages = (None, None)
    if args.stages:
        words = args.stages.split(':')
        try:
            if len(words) == 1:
                stages = (int(words[0]), int(words[0])+1)
            elif len(words) == 2:
                stages = tuple(int(word) if word else None for word in words)
            else:
                raise ValueError()

        except ValueError:
            raise ToolError('Invalid --stages argument.')

    data = []
    for response in squirrel.get_responses(**args.squirrel_query):
        print(response.summary)
        if args.problems:
            for level, message, _ in response.log:
                print('!   %s: %s' % (
                    level.capitalize(),
                    indent('!', 0, 5, message)))

        if args.level == 'stages':
            for stage in response.stages[slice(*stages)]:
                print('  %s' % stage.summary)
                if args.problems:
                    for level, message, _ in stage.log:
                        print('!     %s: %s' % (
                            level.capitalize(),
                            indent('!', 0, 7, message)))

        data.append((
            ', '.join([
                str(response.codes),
                response.str_time_span_short,
                response.summary_log,
                response.summary_quantities]),
            response.get_effective(
                input_quantity=args.input_quantity,
                stages=stages)))

    if args.plot:
        if not data:
            logger.warning('No response objects found.')
            return

        labels, resps = [list(x) for x in zip(*data)]

        from pyrockoeost.plot.response import plot
        plot(
            resps,
            fmin=args.fmin,
            fmax=args.fmax,
            nf=200,
            normalize=args.normalize,
            labels=labels,
            filename=args.out_path,
            dpi=args.dpi,
            show_breakpoints=args.show_breakpoints,
            separate_combined_labels='print' if args.index_labels else None)



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

'''
Low-level FDSN web service client.

This module provides basic functionality to download station metadata, time
series data and event information from FDSN web services. Password and token
authentication are supported. Query responses are returned as open file-like
objects or can be parsed into pyrockoeost's native data structures where
appropriate.

.. _registered-site-names:

Registered site names
.....................

A list of known FDSN site names is maintained within the module for quick
selection by the user. This list currently contains the following sites:

%s

Any other site can be specified by providing its full URL.
'''

import re
import logging
import socket


from pyrockoeost import util
from pyrockoeost.util import DownloadError
from pyrockoeost import config

from pyrockoeost.util import \
    urlencode, Request, build_opener, HTTPDigestAuthHandler, urlopen, HTTPError


logger = logging.getLogger('pyrockoeost.client.fdsn')

g_url = '%(site)s/fdsnws/%(service)s/%(majorversion)i/%(method)s'

g_site_abbr = {
    'geofon': 'https://geofon.gfz-potsdam.de',
    'iris': 'http://service.iris.edu',
    'orfeus': 'http://www.orfeus-eu.org',
    'bgr': 'http://eida.bgr.de',
    'emsc': 'http://www.seismicportal.eu',
    'geonet': 'http://service.geonet.org.nz',
    'knmi': 'http://rdsa.knmi.nl',
    'ncedc': 'http://service.ncedc.org',
    'scedc': 'http://service.scedc.caltech.edu',
    'usgs': 'http://earthquake.usgs.gov',
    'koeri': 'http://eida-service.koeri.boun.edu.tr',
    'ethz': 'http://eida.ethz.ch',
    'icgc': 'http://ws.icgc.cat',
    'ipgp': 'http://eida.ipgp.fr',
    'ingv': 'http://webservices.ingv.it',
    'isc': 'http://www.isc.ac.uk',
    'lmu': 'http://erde.geophysik.uni-muenchen.de',
    'noa': 'http://eida.gein.noa.gr',
    'resif': 'http://ws.resif.fr',
    'usp': 'http://seisrequest.iag.usp.br',
    'niep': 'http://eida-sc3.infp.ro'
}

g_default_site = 'geofon'


g_default_query_args = {
    'station': {
        'starttime', 'endtime', 'startbefore', 'startafter', 'endbefore',
        'endafter', 'network', 'station', 'location', 'channel', 'minlatitude',
        'maxlatitude', 'minlongitude', 'maxlongitude', 'latitude', 'longitude',
        'minradius', 'maxradius', 'level', 'includerestricted',
        'includeavailability', 'updatedafter', 'matchtimeseries', 'format',
        'nodata'},
    'dataselect': {
        'starttime', 'endtime', 'network', 'station', 'location', 'channel',
        'quality', 'minimumlength', 'longestonly', 'format', 'nodata'},
    'event': {
        'starttime', 'endtime', 'minlatitude', 'maxlatitude', 'minlongitude',
        'maxlongitude', 'latitude', 'longitude', 'minradius', 'maxradius',
        'mindepth', 'maxdepth', 'minmagnitude', 'maxmagnitude', 'eventtype',
        'includeallorigins', 'includeallmagnitudes', 'includearrivals',
        'eventid', 'limit', 'offset', 'orderby', 'catalog', 'contributor',
        'updatedafter', 'format', 'nodata'},
    'availability': {
        'starttime', 'endtime', 'network', 'station', 'location', 'channel',
        'quality', 'merge', 'orderby', 'limit', 'includerestricted', 'format',
        'nodata', 'mergegaps', 'show'}}


def doc_escape_slist(li):
    return ', '.join("``'%s'``" % s for s in li)


def doc_table_dict(d, khead, vhead, indent=''):
    keys, vals = zip(*sorted(d.items()))

    lk = max(max(len(k) for k in keys), len(khead))
    lv = max(max(len(v) for v in vals), len(vhead))

    hr = '=' * lk + ' ' + '=' * lv

    lines = [
        hr,
        '%s %s' % (khead.ljust(lk), vhead.ljust(lv)),
        hr]

    for k, v in zip(keys, vals):
        lines.append('%s %s' % (k.ljust(lk), v.ljust(lv)))

    lines.append(hr)
    return '\n'.join(indent + line for line in lines)


def strip_html(s):
    s = s.decode('utf-8')
    s = re.sub(r'<[^>]+>', '', s)
    s = re.sub(r'\r', '', s)
    s = re.sub(r'\s*\n', '\n', s)
    return s


def indent(s, ind='  '):
    return '\n'.join(ind + line for line in s.splitlines())


def get_sites():
    '''
    Get sorted list of registered site names.
    '''
    return sorted(g_site_abbr.keys())


if config.config().fdsn_timeout is None:
    g_timeout = 20.
else:
    g_timeout = config.config().fdsn_timeout

re_realm_from_auth_header = re.compile(r'(realm)\s*[:=]\s*"([^"]*)"?')


class CannotGetRealmFromAuthHeader(DownloadError):
    '''
    Raised when failing to parse server response during authentication.
    '''
    pass


class CannotGetCredentialsFromAuthRequest(DownloadError):
    '''
    Raised when failing to parse server response during token authentication.
    '''
    pass


def get_realm_from_auth_header(headers):
    realm = dict(re_realm_from_auth_header.findall(
        headers['WWW-Authenticate'])).get('realm', None)

    if realm is None:
        raise CannotGetRealmFromAuthHeader('headers=%s' % str(headers))

    return realm


def sdatetime(t):
    return util.time_to_str(t, format='%Y-%m-%dT%H:%M:%S')


class EmptyResult(DownloadError):
    '''
    Raised when an empty server response is retrieved.
    '''
    def __init__(self, url):
        DownloadError.__init__(self)
        self._url = url

    def __str__(self):
        return 'No results for request %s' % self._url


class RequestEntityTooLarge(DownloadError):
    '''
    Raised when the server indicates that too much data was requested.
    '''
    def __init__(self, url):
        DownloadError.__init__(self)
        self._url = url

    def __str__(self):
        return 'Request entity too large: %s' % self._url


class InvalidRequest(DownloadError):
    '''
    Raised when an invalid request would be sent / has been sent.
    '''
    pass


class Timeout(DownloadError):
    '''
    Raised when the server does not respond within the allowed timeout period.
    '''
    pass


def _request(
        url,
        post=False,
        user=None,
        passwd=None,
        timeout=g_timeout,
        **kwargs):

    url_values = urlencode(kwargs)
    if url_values:
        url += '?' + url_values

    logger.debug('Accessing URL %s' % url)
    url_args = {
        'timeout': timeout
    }

    if util.g_ssl_context:
        url_args['context'] = util.g_ssl_context

    opener = None

    req = Request(url)
    if post:
        if isinstance(post, str):
            post = post.encode('utf8')
        logger.debug('POST data: \n%s' % post.decode('utf8'))
        req.data = post

    req.add_header('Accept', '*/*')

    itry = 0
    while True:
        itry += 1
        try:
            urlopen_ = opener.open if opener else urlopen
            try:
                resp = urlopen_(req, **url_args)
            except TypeError:
                # context and cafile not avail before 3.4.3, 2.7.9
                url_args.pop('context', None)
                url_args.pop('cafile', None)
                resp = urlopen_(req, **url_args)

            logger.debug('Response: %s' % resp.getcode())
            if resp.getcode() == 204:
                raise EmptyResult(url)
            return resp

        except HTTPError as e:
            if e.code == 413:
                raise RequestEntityTooLarge(url)

            elif e.code == 401:
                headers = getattr(e, 'headers', e.hdrs)

                realm = get_realm_from_auth_header(headers)

                if itry == 1 and user is not None:
                    auth_handler = HTTPDigestAuthHandler()
                    auth_handler.add_password(
                        realm=realm,
                        uri=url,
                        user=user,
                        passwd=passwd or '')

                    opener = build_opener(auth_handler)
                    continue
                else:
                    raise DownloadError(
                        'Authentication failed for realm "%s" when accessing '
                        'url "%s". Original error was: %s' % (
                            realm, url, str(e)))

            else:
                raise DownloadError(
                    'Error content returned by server (HTML stripped):\n%s\n'
                    '  Original error was: %s' % (
                        indent(
                            strip_html(e.read()),
                            '  !  '),
                        str(e)))

        except socket.timeout:
            raise Timeout(
                'Timeout error. No response received within %i s. You '
                'may want to retry with a longer timeout setting.' % timeout)

        break


def fillurl(service, site, url, majorversion, method):
    return url % dict(
        site=g_site_abbr.get(site, site),
        service=service,
        majorversion=majorversion,
        method=method)


def fix_params(d):

    params = dict(d)
    for k in ['starttime',
              'endtime',
              'startbefore',
              'startafter',
              'endbefore',
              'endafter',
              'updatedafter']:

        if k in params:
            params[k] = sdatetime(params[k])

    if params.get('location', None) == '':
        params['location'] = '--'

    for k in params:
        if isinstance(params[k], bool):
            params[k] = ['false', 'true'][bool(params[k])]

    return params


def make_data_selection(
        stations, tmin, tmax,
        channel_prio=[['BHZ', 'HHZ'],
                      ['BH1', 'BHN', 'HH1', 'HHN'],
                      ['BH2', 'BHE', 'HH2', 'HHE']]):

    selection = []
    for station in stations:
        wanted = []
        for group in channel_prio:
            gchannels = []
            for channel in station.get_channels():
                if channel.name in group:
                    gchannels.append(channel)
            if gchannels:
                gchannels.sort(key=lambda a: group.index(a.name))
                wanted.append(gchannels[0])

        if wanted:
            for channel in wanted:
                selection.append((station.network, station.station,
                                  station.location, channel.name, tmin, tmax))

    return selection


def station(
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout,
        check=True,
        selection=None,
        parsed=True,
        **kwargs):

    '''
    Query FDSN web service for station metadata.

    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional
    :param check:
        If ``True`` arguments are checked against self-description (WADL) of
        the queried web service if available or FDSN specification.
    :type check: bool, optional
    :param selection:
        If given, selection to be queried as a list of tuples
        ``(network, station, location, channel, tmin, tmax)``. Useful for
        detailed queries.
    :type selection: list of tuples, optional
    :param parsed:
        If ``True`` parse received content into
        :py:class:`~pyrockoeost.io.stationxml.FDSNStationXML`
        object, otherwise return open file handle to raw data stream.
    :type parsed: bool, optional
    :param \\*\\*kwargs:
        Parameters passed to the server (see `FDSN web services specification
        <https://www.fdsn.org/webservices>`_).

    :returns:
        See description of ``parsed`` argument above.

    :raises:
        On failure, :py:exc:`~pyrockoeost.util.DownloadError` or one of its
        sub-types defined in the :py:mod:`~pyrockoeost.client.fdsn` module is
        raised.
    '''

    service = 'station'

    if check:
        check_params(service, site, url, majorversion, timeout, **kwargs)

    params = fix_params(kwargs)

    url = fillurl(service, site, url, majorversion, 'query')
    if selection:
        lst = []
        for k, v in params.items():
            lst.append('%s=%s' % (k, v))

        for (network, station, location, channel, tmin, tmax) in selection:
            if location == '':
                location = '--'

            lst.append(' '.join((network, station, location, channel,
                                 sdatetime(tmin), sdatetime(tmax))))

        post = '\n'.join(lst)
        params = dict(post=post.encode())

    if parsed:
        from pyrockoeost.io import stationxml
        format = kwargs.get('format', 'xml')
        if format == 'text':
            if kwargs.get('level', 'station') == 'channel':
                return stationxml.load_channel_table(
                    stream=_request(url, timeout=timeout, **params))
            else:
                raise InvalidRequest('if format="text" shall be parsed, '
                                     'level="channel" is required')

        elif format == 'xml':
            return stationxml.load_xml(
                stream=_request(url, timeout=timeout, **params))
        else:
            raise InvalidRequest('format must be "xml" or "text"')
    else:
        return _request(url, timeout=timeout, **params)


def get_auth_credentials(service, site, url, majorversion, token, timeout):

    url = fillurl(service, site, url, majorversion, 'auth')

    f = _request(url, timeout=timeout, post=token)
    s = f.read().decode()
    try:
        user, passwd = s.strip().split(':')
    except ValueError:
        raise CannotGetCredentialsFromAuthRequest('data="%s"' % s)

    return user, passwd


def dataselect(
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout,
        check=True,
        user=None,
        passwd=None,
        token=None,
        selection=None,
        **kwargs):

    '''
    Query FDSN web service for time series data in miniSEED format.

    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional
    :param check:
        If ``True`` arguments are checked against self-description (WADL) of
        the queried web service if available or FDSN specification.
    :type check: bool, optional
    :param user: User name for user/password authentication.
    :type user: str, optional
    :param passwd: Password for user/password authentication.
    :type passwd: str, optional
    :param token: Token for `token authentication
        <https://geofon.gfz-potsdam.de/waveform/archive/auth/auth-overview.php>`_.
    :type token: str, optional
    :param selection:
        If given, selection to be queried as a list of tuples
        ``(network, station, location, channel, tmin, tmax)``.
    :type selection: list of tuples, optional
    :param \\*\\*kwargs:
        Parameters passed to the server (see `FDSN web services specification
        <https://www.fdsn.org/webservices>`_).

    :returns:
        Open file-like object providing raw miniSEED data.
    '''

    service = 'dataselect'

    if user or token:
        method = 'queryauth'
    else:
        method = 'query'

    if token is not None:
        user, passwd = get_auth_credentials(
            service, site, url, majorversion, token, timeout)

    if check:
        check_params(service, site, url, majorversion, timeout, **kwargs)

    params = fix_params(kwargs)

    url = fillurl(service, site, url, majorversion, method)
    if selection:
        lst = []

        for k, v in params.items():
            lst.append('%s=%s' % (k, v))

        for (network, station, location, channel, tmin, tmax) in selection:
            if location == '':
                location = '--'

            lst.append(' '.join((network, station, location, channel,
                                 sdatetime(tmin), sdatetime(tmax))))

        post = '\n'.join(lst)
        return _request(
            url, user=user, passwd=passwd, post=post.encode(), timeout=timeout)
    else:
        return _request(
            url, user=user, passwd=passwd, timeout=timeout, **params)


def event(
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout,
        check=True,
        user=None,
        passwd=None,
        token=None,
        parsed=False,
        **kwargs):

    '''
    Query FDSN web service for events.

    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional
    :param check:
        If ``True`` arguments are checked against self-description (WADL) of
        the queried web service if available or FDSN specification.
    :type check: bool, optional
    :param user: User name for user/password authentication.
    :type user: str, optional
    :param passwd: Password for user/password authentication.
    :type passwd: str, optional
    :param token: Token for `token authentication
        <https://geofon.gfz-potsdam.de/waveform/archive/auth/auth-overview.php>`_.
    :type token: str, optional
    :param parsed:
        If ``True`` parse received content into
        :py:class:`~pyrockoeost.io.quakeml.QuakeML`
        object, otherwise return open file handle to raw data stream. Note:
        by default unparsed data is retrieved, differently from the default
        behaviour of :py:func:`station` (for backward compatibility).
    :type parsed: bool, optional
    :param \\*\\*kwargs:
        Parameters passed to the server (see `FDSN web services specification
        <https://www.fdsn.org/webservices>`_).

    :returns:
        See description of ``parsed`` argument above.
    '''

    service = 'event'

    if user or token:
        method = 'queryauth'
    else:
        method = 'query'

    if token is not None:
        user, passwd = get_auth_credentials(
            service, site, url, majorversion, token, timeout)

    if check:
        check_params(service, site, url, majorversion, timeout, **kwargs)

    params = fix_params(kwargs)

    url = fillurl(service, site, url, majorversion, method)

    fh = _request(url, user=user, passwd=passwd, timeout=timeout, **params)
    if parsed:
        from pyrockoeost.io import quakeml
        format = kwargs.get('format', 'xml')
        if format != 'xml':
            raise InvalidRequest(
                'If parsed=True is selected, format="xml" must be selected.')

        return quakeml.QuakeML.load_xml(stream=fh)

    else:
        return fh


def availability(
        method='query',
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout,
        check=True,
        user=None,
        passwd=None,
        token=None,
        selection=None,
        **kwargs):

    '''
    Query FDSN web service for time series data availablity.

    :param method: Availablility method to call: ``'query'``, or ``'extent'``.
    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional
    :param check:
        If ``True`` arguments are checked against self-description (WADL) of
        the queried web service if available or FDSN specification.
    :type check: bool, optional
    :param user: User name for user/password authentication.
    :type user: str, optional
    :param passwd: Password for user/password authentication.
    :type passwd: str, optional
    :param token: Token for `token authentication
        <https://geofon.gfz-potsdam.de/waveform/archive/auth/auth-overview.php>`_.
    :type token: str, optional
    :param selection:
        If given, selection to be queried as a list of tuples
        ``(network, station, location, channel, tmin, tmax)``.
    :type selection: list of tuples, optional
    :param \\*\\*kwargs:
        Parameters passed to the server (see `FDSN web services specification
        <https://www.fdsn.org/webservices>`_).

    :returns:
        Open file-like object providing raw response.
    '''

    service = 'availability'

    assert method in ('query', 'extent')

    if user or token:
        method += 'auth'

    if token is not None:
        user, passwd = get_auth_credentials(
            service, site, url, majorversion, token, timeout)

    if check:
        check_params(service, site, url, majorversion, timeout, **kwargs)

    params = fix_params(kwargs)

    url = fillurl(service, site, url, majorversion, method)
    if selection:
        lst = []

        for k, v in params.items():
            lst.append('%s=%s' % (k, v))

        for (network, station, location, channel, tmin, tmax) in selection:
            if location == '':
                location = '--'

            lst.append(' '.join((network, station, location, channel,
                                 sdatetime(tmin), sdatetime(tmax))))

        post = '\n'.join(lst)
        return _request(
            url, user=user, passwd=passwd, post=post.encode(), timeout=timeout)
    else:
        return _request(
            url, user=user, passwd=passwd, timeout=timeout, **params)


def check_params(
        service,
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout,
        method='query',
        **kwargs):

    '''
    Check query parameters against self-description of web service.

    Downloads WADL description of the given service and site and checks
    parameters if they are available. Queried WADLs are cached in memory.

    :param service: ``'station'``, ``'dataselect'``, ``'event'`` or
        ``'availability'``.
    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional
    :param \\*\\*kwargs:
        Parameters that would be passed to the server (see `FDSN web services
        specification <https://www.fdsn.org/webservices>`_).

    :raises: :py:exc:`ValueError` is raised if unsupported parameters are
        encountered.
    '''

    avail = supported_params_wadl(
        service, site, url, majorversion, timeout, method)

    unavail = sorted(set(kwargs.keys()) - avail)
    if unavail:
        raise ValueError(
            'Unsupported parameter%s for service "%s" at site "%s": %s' % (
                '' if len(unavail) == 1 else 's',
                service,
                site,
                ', '.join(unavail)))


def supported_params_wadl(
        service,
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout,
        method='query'):

    '''
    Get query parameter names supported by a given FDSN site and service.

    If no WADL is provided by the queried service, default parameter sets from
    the FDSN standard are returned. Queried WADLs are cached in memory.

    :param service: ``'station'``, ``'dataselect'``, ``'event'`` or
        ``'availability'``.
    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional

    :returns: Supported parameter names.
    :rtype: set of str
    '''

    wadl = cached_wadl(service, site, url, majorversion, timeout)

    if wadl:
        url = fillurl(service, site, url, majorversion, method)
        return set(wadl.supported_param_names(url))
    else:
        return g_default_query_args[service]


def patch_geonet_wadl(wadl):
    for r in wadl.resources_list:
        r.base = r.base.replace('1/station', 'station/1')
        r.base = r.base.replace('1/dataselect', 'dataselect/1')
        r.base = r.base.replace('1/event', 'event/1')


def wadl(
        service,
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout):

    '''
    Retrieve self-description of a specific FDSN service.

    :param service: ``'station'``, ``'dataselect'``, ``'event'`` or
        ``'availability'``.
    :param site:
        :ref:`Registered site name <registered-site-names>` or full base URL of
        the service (e.g. ``'https://geofon.gfz-potsdam.de'``).
    :type site: str, optional
    :param url:
        URL template (default should work in 99% of cases).
    :type url: str, optional
    :param majorversion:
        Major version of the service to query (always ``1`` at the time of
        writing).
    :type majorversion: int, optional
    :param timeout:
        Network timeout in [s]. Global default timeout can be configured in
        pyrockoeost's configuration file under ``fdsn_timeout``.
    :type timeout: float, optional
    '''

    from pyrockoeost.client.wadl import load_xml

    url = fillurl(service, site, url, majorversion, 'application.wadl')

    wadl = load_xml(stream=_request(url, timeout=timeout))

    if site == 'geonet' or site.find('geonet.org.nz') != -1:
        patch_geonet_wadl(wadl)

    return wadl


g_wadls = {}


def cached_wadl(
        service,
        site=g_default_site,
        url=g_url,
        majorversion=1,
        timeout=g_timeout):

    '''
    Get self-description of a specific FDSN service.

    Same as :py:func:`wadl`, but results are cached in memory.
    '''

    k = (service, site, url, majorversion)
    if k not in g_wadls:
        try:
            g_wadls[k] = wadl(service, site, url, majorversion, timeout)

        except Timeout:
            raise

        except DownloadError:
            logger.info(
                'No service description (WADL) found for "%s" at site "%s".'
                % (service, site))

            g_wadls[k] = None

    return g_wadls[k]


__doc__ %= doc_table_dict(g_site_abbr, 'Site name', 'URL', '    ')


if __name__ == '__main__':
    import sys

    util.setup_logging('pyrockoeost.client.fdsn', 'info')

    if len(sys.argv) == 1:
        sites = get_sites()
    else:
        sites = sys.argv[1:]

    for site in sites:
        print('=== %s (%s) ===' % (site, g_site_abbr[site]))

        for service in ['station', 'dataselect', 'event']:
            try:
                app = wadl(service, site=site, timeout=2.0)
                print(indent(str(app)))

            except Timeout as e:
                logger.error(str(e))
                print('%s: timeout' % (service,))

            except util.DownloadError as e:
                logger.error(str(e))
                print('%s: no wadl' % (service,))



import time
import logging
import numpy as num
from pyrockoeost import util, trace

logger = logging.getLogger('pyrockoeost.streaming.datacube')

cube_rates = num.array([50., 100., 200., 400., 800.], dtype=float)


def quotechars(chars):
    return ''.join(['.', chr(c)][chr(c).isalnum()] for c in chars)


def hexdump(chars, sep=' ', width=16):
    while chars:
        line = chars[:width]
        chars = chars[width:]
        line = line.ljust(width, b'\000')

        print('%s%s%s' % (
            sep.join('%02x' % c for c in line),
            sep, quotechars(line)))


def read_wait(s, nread):
    buf = b''
    while True:
        buf += s.read(nread - len(buf))
        if len(buf) == nread:
            return buf

        logger.info('Incomplete read from serial port. Waiting...')
        time.sleep(1.0)


def read_blocktype(s):
    buf = read_wait(s, 1)
    return buf[0] >> 4


def determine_nchannels(s):
    while True:
        datablock_starts = []
        for i in range(1024):
            datablock_starts.append(read_blocktype(s) in (8, 9))

        if len(datablock_starts) > 50:
            dd = num.diff(num.where(datablock_starts)[0]) - 1
            dd = dd[dd % 4 == 0]

            if dd.size > 50:
                nchannels = dd // 4
                nchannels_hist, _ = num.histogram(nchannels, num.arange(5)-1)
                nchannels = num.argmax(nchannels_hist)
                if nchannels_hist[nchannels] > 50:
                    return nchannels

        logger.info('Could not determine number of channels. Retrying...')


def sync(s, nchannels, nok_want=10):
    nok = 0
    while nok < nok_want:
        blocktype = read_blocktype(s)
        if blocktype in (8, 9):
            read_wait(s, 4*nchannels)
            nok += 1
        elif blocktype == 10:
            read_wait(s, 79)
            nok += 1
        else:
            nok = 0


def read_sample(s, nchannels):

    b = read_wait(s, nchannels*4)

    sample = []

    for i in range(nchannels):
        v = b[i*4 + 0] << 17
        v += b[i*4 + 1] << 10
        v += b[i*4 + 2] << 3
        v += b[i*4 + 3]
        v -= (v & 0x800000) << 1
        sample.append(v)

    return sample


class SerialCubeError(Exception):
    pass


class SerialCube(object):

    def __init__(
            self,
            device='/dev/ttyUSB0',
            network='',
            station='CUBE',
            location='',
            timeout=5.0):

        # fixed
        self._network = network
        self._station = station
        self._location = location
        self._device = device
        self._timeout = timeout
        self._baudrate = 115200
        self._nsamples_buffer = 100

        # state
        self._init_state()

    def _init_state(self):
        self._serial = None
        self._nchannels = None
        self._tstart = None
        self._rate = None
        self._buffer = None
        self._isamples_buffer = None
        self._nsamples_read = 0

    def acquisition_start(self):
        import serial
        assert self._serial is None
        try:
            self._serial = serial.Serial(
                port=self._device,
                baudrate=self._baudrate,
                timeout=self._timeout)

        except serial.serialutil.SerialException as e:
            raise SerialCubeError(
                'Opening serial interface failed: %s' % str(e))

        nchannels = determine_nchannels(self._serial)
        logger.info('Number of channels: %i' % nchannels)

        sync(self._serial, nchannels)
        self._nchannels = nchannels

    def acquisition_stop(self):
        assert self._serial is not None
        self._serial.close()
        self._init_state()

    def process(self):
        assert self._serial is not None

        blocktype = read_blocktype(self._serial)
        if blocktype in (8, 9):
            if self._buffer is None:
                self._buffer = num.zeros(
                    (self._nsamples_buffer, self._nchannels),
                    dtype=num.float32)
                if self._tstart is None:
                    self._tstart = time.time()

                self._isamples_buffer = 0

            self._buffer[self._isamples_buffer, :] = read_sample(
                self._serial, self._nchannels)

            self._isamples_buffer += 1

        elif blocktype == 10:
            read_wait(self._serial, 79)

        if self._isamples_buffer == self._nsamples_buffer:
            if self._rate is None:
                tdur = time.time() - self._tstart
                assert tdur > 0.
                rate_approx = self._nsamples_buffer / tdur
                self._rate = cube_rates[
                    num.argmin(num.abs(rate_approx - cube_rates))]

                logger.info('Sample rate [Hz]: %g' % self._rate)

            deltat = 1.0 / self._rate
            t = self._tstart + self._nsamples_read * deltat
            for ichannel in range(self._nchannels):
                tr = trace.Trace(
                    self._network,
                    self._station,
                    self._location,
                    'p%i' % ichannel,
                    ydata=self._buffer[:, ichannel],
                    tmin=t,
                    deltat=1.0/self._rate)

                self.got_trace(tr)

            self._nsamples_read += self._nsamples_buffer
            self._buffer = None
            self._isamples_buffer = None

    def got_trace(self, tr):
        logger.info(
            'Got trace from DATA-CUBE: %s, mean: %g, std: %g' % (
                tr.summary, num.mean(tr.ydata), num.std(tr.ydata)))


def main():
    import sys
    util.setup_logging('main', 'info')
    if len(sys.argv) != 2:
        sys.exit('usage: python -m pyrockoeost.streaming.datacube DEVICE')

    device = sys.argv[1]
    cs = SerialCube(device)
    try:
        cs.acquisition_start()
    except SerialCubeError as e:
        sys.exit(str(e))

    try:
        while True:
            cs.process()

    except KeyboardInterrupt:
        pass

    finally:
        cs.acquisition_stop()


if __name__ == '__main__':
    main()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from pyrockoeost.squirrel.model import to_kind


headline = 'Get summary of available station/channel codes.'


def make_subparser(subparsers):
    return subparsers.add_parser(
        'codes',
        help=headline,
        description=headline)


def setup(parser):
    parser.add_squirrel_selection_arguments()
    parser.add_squirrel_query_arguments(without=['time', 'tmin', 'tmax'])


def run(parser, args):
    from pyrockoeost import squirrel as sq

    squirrel = args.make_squirrel()

    kwargs = args.squirrel_query
    kinds = kwargs.pop('kind', sq.supported_content_kinds())
    codes_query = kwargs.pop('codes', None)

    for kind in kinds:
        for kind_id, codes, deltat, _, count in sorted(
                squirrel._iter_codes_info(kind=kind, codes=codes_query)):
            print(to_kind(kind_id), codes, deltat, count)



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import string

import numpy as num

from pyrockoeost.guts import Bool, Float, Object, String

from pyrockoeost import cake, geometry, gf
from pyrockoeost.gui.qt_compat import qc, qw
from pyrockoeost.gui.talkie import TalkieRoot

from pyrockoeost.gui.vtk_util import \
    ArrowPipe, ColorbarPipe, PolygonPipe, ScatterPipe, OutlinesPipe

from .. import state as vstate
from .. import common
from . import base

guts_prefix = 'sparrow'


d2r = num.pi / 180.


map_anchor = {
    'center': (0.0, 0.0),
    'center_left': (-1.0, 0.0),
    'center_right': (1.0, 0.0),
    'top': (0.0, -1.0),
    'top_left': (-1.0, -1.0),
    'top_right': (1.0, -1.0),
    'bottom': (0.0, 1.0),
    'bottom_left': (-1.0, 1.0),
    'bottom_right': (1.0, 1.0)}


class ProxySource(TalkieRoot):
    pass


for source_cls in [gf.RectangularSource]:

    cls_name = 'Proxy' + source_cls.__name__

    class proxy_source_cls(ProxySource):
        class_name = cls_name

        def __init__(self, **kwargs):
            ProxySource.__init__(self)
            for key, value in self._ranges.items():
                setattr(self, key, value['ini'])

            if kwargs is not None:
                for it in kwargs.items():
                    setattr(self, it[0], it[1])

    proxy_source_cls.__name__ = cls_name
    vars()[cls_name] = proxy_source_cls

    for prop in source_cls.T.properties:
        proxy_source_cls.T.add_property(prop.name, prop)

ProxyRectangularSource = vars()['ProxyRectangularSource']  # silence flake8

ProxyRectangularSource._name = 'RectangularSource'

ProxyRectangularSource._ranges = {
    'lat': {'min': -90., 'max': 90., 'step': 1, 'ini': 0.},
    'lon': {'min': -180., 'max': 180., 'step': 1, 'ini': 0.},
    'depth': {'min': 0., 'max': 600000., 'step': 1000, 'ini': 10000.},
    'width': {'min': 0.1, 'max': 500000., 'step': 1000, 'ini': 10000.},
    'length': {'min': 0.1, 'max': 1000000., 'step': 1000, 'ini': 50000.},
    'strike': {'min': -180., 'max': 180., 'step': 1, 'ini': 0.},
    'dip': {'min': 0., 'max': 90., 'step': 1, 'ini': 45.},
    'rake': {'min': -180., 'max': 180., 'step': 1, 'ini': 0.},
    'nucleation_x':
        {'min': -100., 'max': 100., 'step': 1, 'ini': 0., 'fac': .01},
    'nucleation_y':
        {'min': -100., 'max': 100., 'step': 1, 'ini': 0., 'fac': .01},
    'slip': {'min': 0., 'max': 1000., 'step': 1, 'ini': 1., 'fac': .01}}


class ProxyConfig(Object):
    deltas = num.array([1000., 1000.])
    deltat = Float.T(default=0.5)
    rho = Float.T(default=2800)
    vs = Float.T(default=3600)

    def get_shear_moduli(self, *args, **kwargs):
        points = kwargs.get('points')
        return num.ones(len(points)) * num.power(self.vs, 2) * self.rho


class ProxyStore(Object):
    def __init__(self, **kwargs):
        config = ProxyConfig()
        if kwargs:
            config.deltas = kwargs.get('deltas', config.deltas)
            config.deltat = kwargs.get('deltat', config.deltat)
            config.rho = kwargs.get('rho', config.rho)
            config.vs = kwargs.get('vs', config.vs)

        self.config = config
        self.mode = String.T(default='r')
        self._f_data = None
        self._f_index = None


parameter_label = {
    'time (s)': 'times',
    'slip (m)': 'slip',
    'moment (Nm)': 'moment'
}

parameter_geometry = {
    'time (s)': 't_arrival',
    'slip (m)': 'slip',
    'moment (Nm)': 'moment'
}

unit_label = {
    'lat': '(deg)',
    'lon': '(deg)',
    'depth': '(m)',
    'strike': '(deg)',
    'dip': '(deg)',
    'rake': '(deg)',
    'length': '(m)',
    'width': '(m)',
    'slip': '(m)'
}


class SourceState(base.ElementState):
    visible = Bool.T(default=True)
    source_selection = ProxySource.T(default=ProxyRectangularSource())  # noqa
    deltat = Float.T(default=0.5)
    display_parameter = String.T(default='time (s)')
    cpt = base.CPTState.T(default=base.CPTState.D())

    @classmethod
    def get_name(self):
        return 'Rectangular Source'

    def create(self):
        element = SourceElement()
        return element


class SourceElement(base.Element):

    def __init__(self):
        base.Element.__init__(self)
        self._parent = None
        self._pipe = []
        self._controls = None
        self._points = num.array([])

        self.cpt_handler = base.CPTHandler()

    def _state_bind_source(self, *args, **kwargs):
        vstate.state_bind(self, self._state.source_selection, *args, **kwargs)

    def _state_bind_store(self, *args, **kwargs):
        vstate.state_bind(self, self._state, *args, **kwargs)

    def bind_state(self, state):
        base.Element.bind_state(self, state)
        self.talkie_connect(
            state,
            ['visible', 'source_selection', 'deltat', 'display_parameter'],
            self.update)

        self.cpt_handler.bind_state(state.cpt, self.update)

    def unbind_state(self):
        self.cpt_handler.unbind_state()
        base.Element.unbind_state(self)

    def get_name(self):
        return 'Rectangular Source'

    def set_parent(self, parent):
        self._parent = parent
        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.update()

    def unset_parent(self):
        self.unbind_state()
        if self._parent:
            if self._pipe:
                for pipe in self._pipe:
                    if isinstance(pipe.actor, list):
                        for act in pipe.actor:
                            self._parent.remove_actor(act)
                    else:
                        self._parent.remove_actor(pipe.actor)
                self._pipe = []

            if self._controls:
                self._parent.remove_panel(self._controls)
                self._controls = None

            self._parent.update_view()
            self._parent = None

    def open_file_load_dialog(self):
        caption = 'Select one file to open'
        fns, _ = qw.QFileDialog.getOpenFileNames(
            self._parent, caption, options=common.qfiledialog_options)

        if fns:
            try:
                self.load_source_file(str(fns[0]))
            except FileNotFoundError as e:
                raise e

        else:
            return

    def load_source_file(self, path):
        loaded_source = gf.load(filename=path)
        source = ProxyRectangularSource(
            **{prop: getattr(loaded_source, prop)
                for prop in loaded_source.T.propnames
                if getattr(loaded_source, prop)})

        self._parent.remove_panel(self._controls)
        self._controls = None
        self._state.source_selection = source
        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.update()

    def open_file_save_dialog(self, fn=None):
        caption = 'Choose a file name to write source'
        if not fn:
            fn, _ = qw.QFileDialog.getSaveFileName(
                self._parent, caption, options=common.qfiledialog_options)
        if fn:
            self.save_file(str(fn))

    def save_file(self, path):
        source = self._state.source_selection
        source2dump = gf.RectangularSource(
            **{prop: getattr(source, prop) for prop in source.T.propnames})

        if path.split('.')[-1].lower() in ['xml']:
            source2dump.dump_xml(filename=path)
        else:
            source2dump.dump(filename=path)

    def update_loc(self, *args):
        pstate = self._parent.state
        state = self._state

        source = state.source_selection
        source.lat = pstate.lat
        source.lon = pstate.lon

        self._state.source_selection.source = source

        self.update()

    def update_source(self, store):
        state = self._state

        source = state.source_selection
        source_list = gf.source_classes

        for i, a in enumerate(source_list):
            if a.__name__ is source._name:
                fault = a(
                    **{prop: source.__dict__[prop]
                        for prop in source.T.propnames})

                source_geom = fault.geometry(store)

                self._update_outlines(source_geom)
                self._update_scatter(source, fault)
                self._update_raster(source_geom, state.display_parameter)
                self._update_rake_arrow(fault)

    def _update_outlines(self, source_geom):

        if source_geom.outlines:
            self._pipe.append(OutlinesPipe(
                source_geom, color=(1., 1., 1.), cs='latlondepth'))
            self._parent.add_actor(
                self._pipe[-1].actor)

            self._pipe.append(OutlinesPipe(
                source_geom, color=(0.6, 0.6, 0.6), cs='latlon'))
            self._parent.add_actor(
                self._pipe[-1].actor)

    def _update_scatter(self, source, fault):
        for point, color in zip(
                ((source.nucleation_x,
                  source.nucleation_y),
                 map_anchor[source.anchor]),
                (num.array([[1., 0., 0.]]),
                 num.array([[0., 0., 1.]]))):

            points = geometry.latlondepth2xyz(
                fault.xy_to_coord(
                    x=[point[0]], y=[point[1]],
                    cs='latlondepth'),
                planetradius=cake.earthradius)

            vertices = geometry.arr_vertices(points)
            p = ScatterPipe(vertices)
            p.set_symbol('sphere')
            p.set_colors(color)
            self._pipe.append(p)
            self._parent.add_actor(p.actor)

    def _update_raster(self, source_geom, param):
        vertices = geometry.arr_vertices(
            source_geom.get_vertices(col='xyz'))

        faces = source_geom.get_faces()

        if param not in parameter_label:
            raise NameError('No parameter label given for %s', param)

        if not source_geom.has_property(parameter_geometry[param]):
            raise AttributeError(
                'No property within source geometry called %s',
                parameter_geometry[param])

        self.cpt_handler._values = source_geom.get_property(
            parameter_geometry[param])
        cbar_title = parameter_label[param]

        self.cpt_handler.update_cpt()

        poly_pipe = PolygonPipe(
            vertices, faces,
            values=self.cpt_handler._values,
            lut=self.cpt_handler._lookuptable)

        if not source_geom.has_property(parameter_geometry[param]):
            raise AttributeError(
                'No property within source geometry called %s',
                parameter_geometry[param])

        tmin = self._parent.state.tmin_effective
        tmax = self._parent.state.tmax_effective

        times = source_geom.get_property(parameter_geometry[param])
        times += self._state.source_selection.time

        if tmin is not None:
            m1 = times < tmin
        else:
            m1 = num.zeros(times.size, dtype=bool)

        if tmax is not None:
            m3 = tmax < times
        else:
            m3 = num.zeros(times.size, dtype=bool)

        m2 = num.logical_not(num.logical_or(m1, m3))

        if not any(m2):
            poly_pipe.set_alpha(0.)
        # print(m2.astype(num.float_))
        # poly_pipe.set_alpha(m2.copy().astype(num.float_))
        # print(self.cpt_handler._lookuptable.GetRange())

        self._pipe.append(poly_pipe)
        self._parent.add_actor(self._pipe[-1].actor)

        if cbar_title is not None:
            cbar_pipe = ColorbarPipe(
                parent_pipe=poly_pipe, cbar_title=cbar_title,
                lut=self.cpt_handler._lookuptable)

            self._pipe.append(cbar_pipe)
            self._parent.add_actor(self._pipe[-1].actor)

    def _update_rake_arrow(self, fault):
        source = self._state.source_selection
        rake = source.rake * d2r

        nucl_x = source.nucleation_x
        nucl_y = source.nucleation_y

        wd_ln = source.width / source.length

        endpoint = [None] * 2
        endpoint[0] = nucl_x + num.cos(rake) * wd_ln
        endpoint[1] = nucl_y + num.sin(-rake)

        points = geometry.latlondepth2xyz(
            fault.xy_to_coord(
                x=[nucl_x, endpoint[0]],
                y=[nucl_y, endpoint[1]],
                cs='latlondepth'),
            planetradius=cake.earthradius)
        vertices = geometry.arr_vertices(points)

        self._pipe.append(ArrowPipe(vertices[0], vertices[1]))
        self._parent.add_actor(self._pipe[-1].actor)

    def update(self, *args):
        state = self._state

        store = ProxyStore(
            deltat=state.deltat)
        store.config.deltas = num.array(
            [(store.config.deltat * store.config.vs) + 1] * 2)

        if self._pipe:
            for pipe in self._pipe:
                self._parent.remove_actor(pipe.actor)

            self._pipe = []

        if state.visible:
            self.update_source(store)

        self._parent.update_view()

    def _get_controls(self):
        if not self._controls:
            from ..state import \
                state_bind_slider, state_bind_combobox
            from pyrockoeost import gf

            source = self._state.source_selection

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            frame.setLayout(layout)

            def state_to_lineedit(state, attribute, widget):
                sel = getattr(state, attribute)

                widget.setText('%g' % sel)

            def lineedit_to_state(widget, state, attribute):
                s = float(widget.text())
                try:
                    setattr(state, attribute, s)
                except Exception:
                    raise ValueError(
                        'Value of %s needs to be a float or integer'
                        % string.capwords(attribute))

            il = 0

            # Origin time controls
            layout.addWidget(qw.QLabel('Origin time'), il, 0)
            le_time = qw.QLineEdit()
            layout.addWidget(le_time, il, 1, 1, 2)

            self._state_bind_source(
                ['time'], common.lineedit_to_time, le_time,
                [le_time.editingFinished, le_time.returnPressed],
                common.time_to_lineedit,
                attribute='time')

            for var in ['tmin', 'tmax', 'tduration', 'tposition']:
                self.talkie_connect(
                    self._parent.state, var, self.update)

            # Source property controls
            for il, label in enumerate(source.T.propnames, start=il+1):
                if label in source._ranges.keys():

                    unit = unit_label[label] if label in unit_label else ''

                    layout.addWidget(qw.QLabel(
                        f'{string.capwords(label)} {unit}'), il, 0)

                    slider = qw.QSlider(qc.Qt.Horizontal)
                    slider.setSizePolicy(
                        qw.QSizePolicy(
                            qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
                    slider.setMinimum(
                        int(round(source._ranges[label]['min'])))
                    slider.setMaximum(
                        int(round(source._ranges[label]['max'])))
                    slider.setSingleStep(
                        int(round(source._ranges[label]['step'])))
                    slider.setPageStep(
                        int(round(source._ranges[label]['step'])))

                    layout.addWidget(slider, il, 1)
                    try:
                        state_bind_slider(
                            self, self._state.source_selection, label, slider,
                            factor=source._ranges[label]['fac'])
                    except Exception:
                        state_bind_slider(
                            self, self._state.source_selection, label, slider)

                    le = qw.QLineEdit()
                    layout.addWidget(le, il, 2)

                    self._state_bind_source(
                        [label], lineedit_to_state, le,
                        [le.editingFinished, le.returnPressed],
                        state_to_lineedit, attribute=label)

            for label, name in zip(
                    ['Sampling int. (s)'], ['deltat']):
                il += 1
                layout.addWidget(qw.QLabel(label), il, 0)
                slider = qw.QSlider(qc.Qt.Horizontal)
                slider.setSizePolicy(
                    qw.QSizePolicy(
                        qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
                slider.setMinimum(1)
                slider.setMaximum(1000)
                slider.setSingleStep(1)
                slider.setPageStep(1)
                layout.addWidget(slider, il, 1)
                state_bind_slider(
                    self, self._state, name, slider, factor=0.01)

                le = qw.QLineEdit()
                layout.addWidget(le, il, 2)

                self._state_bind_store(
                    [name], lineedit_to_state, le,
                    [le.editingFinished, le.returnPressed],
                    state_to_lineedit, attribute=name)

            il += 1
            layout.addWidget(qw.QLabel('Anchor'), il, 0)

            cb = qw.QComboBox()
            for i, s in enumerate(gf.RectangularSource.anchor.choices):
                cb.insertItem(i, s)
            layout.addWidget(cb, il, 1, 1, 2)
            state_bind_combobox(
                self, self._state.source_selection, 'anchor', cb)

            il += 1
            layout.addWidget(qw.QLabel('Display Param.'), il, 0)

            cb = qw.QComboBox()
            for i, s in enumerate(parameter_label.keys()):
                cb.insertItem(i, s)
            layout.addWidget(cb, il, 1)
            state_bind_combobox(
                self, self._state, 'display_parameter', cb)

            self.cpt_handler.cpt_controls(
                self._parent, self._state.cpt, layout)

            il = layout.rowCount() + 1
            pb = qw.QPushButton('Move Here')
            layout.addWidget(pb, il, 0)
            pb.clicked.connect(self.update_loc)

            pb = qw.QPushButton('Load')
            layout.addWidget(pb, il, 1)
            pb.clicked.connect(self.open_file_load_dialog)

            pb = qw.QPushButton('Save')
            layout.addWidget(pb, il, 2)
            pb.clicked.connect(self.open_file_save_dialog)

            il += 1
            layout.addWidget(qw.QFrame(), il, 0, 1, 3)

        self._controls = frame

        self.cpt_handler._update_cpt_combobox()
        self.cpt_handler._update_cptscale_lineedit()

        return self._controls


__all__ = [
    'SourceElement',
    'SourceState',
]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from .content import *  # noqa
from .location import *  # noqa
from .station import *  # noqa
from .event import *  # noqa
from .gnss import *  # noqa
from .geometry import *  # noqa

from pyrockoeost.util import parse_md

__doc__ = parse_md(__file__)



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import math
import numpy as num
import scipy.signal

from pyrockoeost.orthodrome import positive_region

km = 1e3


class OutOfBounds(Exception):
    pass


class Tile(object):

    def __init__(self, xmin, ymin, dx, dy, data):
        self.xmin = float(xmin)
        self.ymin = float(ymin)
        self.dx = float(dx)
        self.dy = float(dy)
        self.data = data
        self._set_maxes()

    def _set_maxes(self):
        self.ny, self.nx = self.data.shape
        self.xmax = self.xmin + (self.nx-1) * self.dx
        self.ymax = self.ymin + (self.ny-1) * self.dy

    def x(self):
        return self.xmin + num.arange(self.nx) * self.dx

    def y(self):
        return self.ymin + num.arange(self.ny) * self.dy

    def decimate(self, ndeci):
        assert ndeci % 2 == 0
        kernel = num.ones((ndeci+1, ndeci+1))
        kernel /= num.sum(kernel)
        data = scipy.signal.convolve2d(
            self.data.astype(float), kernel, mode='valid')

        self.data = data[::ndeci, ::ndeci].astype(self.data.dtype)
        self.xmin += ndeci/2
        self.ymin += ndeci/2
        self.dx *= ndeci
        self.dy *= ndeci
        self._set_maxes()

    def yextend_with_repeat(self, ymin, ymax):
        assert ymax >= self.ymax
        assert ymin <= self.ymin

        nlo = int(round((self.ymin - ymin) / self.dy))
        nhi = int(round((ymax - self.ymax) / self.dy))

        nx, ny = self.nx, self.ny
        data = num.zeros((ny+nlo+nhi, nx), dtype=self.data.dtype)
        data[:nlo, :] = self.data[nlo, :]
        data[nlo:nlo+ny, :] = self.data
        data[nlo+ny:, :] = self.data[-1, :]

        self.ymin = ymin
        self.data = data
        self._set_maxes()

    def get(self, x, y):
        ix = int(round((x - self.xmin) / self.dx))
        iy = int(round((y - self.ymin) / self.dy))
        if 0 <= ix < self.nx and 0 <= iy < self.ny:
            return self.data[iy, ix]
        else:
            raise OutOfBounds()

    @classmethod
    def from_grd(cls, path):
        '''Load from GMT .grd'''
        from pyrockoeost.plot import gmtpy
        lon, lat, z = gmtpy.loadgrd(path)
        return cls(lon[0], lat[0], lon[1] - lon[0], lat[1] - lat[0],  z)


def multiple_of(x, dx, eps=1e-5):
    return abs(int(round(x / dx))*dx - x) < dx * eps


def combine(tiles, region=None):
    if not tiles:
        return None

    dx = tiles[0].dx
    dy = tiles[0].dy
    dtype = tiles[0].data.dtype

    assert all(t.dx == dx for t in tiles)
    assert all(t.dy == dy for t in tiles)
    assert all(t.data.dtype == dtype for t in tiles)
    assert all(multiple_of(t.xmin, dx) for t in tiles)
    assert all(multiple_of(t.ymin, dy) for t in tiles)

    if region is None:
        xmin = min(t.xmin for t in tiles)
        xmax = max(t.xmax for t in tiles)
        ymin = min(t.ymin for t in tiles)
        ymax = max(t.ymax for t in tiles)
    else:
        xmin, xmax, ymin, ymax = positive_region(region)

        if not multiple_of(xmin, dx):
            xmin = math.floor(xmin / dx) * dx
        if not multiple_of(xmax, dx):
            xmax = math.ceil(xmax / dx) * dx
        if not multiple_of(ymin, dy):
            ymin = math.floor(ymin / dy) * dy
        if not multiple_of(ymax, dy):
            ymax = math.ceil(ymax / dy) * dy

    nx = int(round((xmax - xmin) / dx)) + 1
    ny = int(round((ymax - ymin) / dy)) + 1

    data = num.zeros((ny, nx), dtype=dtype)
    data[:, :] = 0

    for t in tiles:
        for txmin in (t.xmin, t.xmin + 360.):
            ix = int(round((txmin - xmin) / dx))
            iy = int(round((t.ymin - ymin) / dy))
            ixlo = max(ix, 0)
            ixhi = min(ix+t.nx, nx)
            iylo = max(iy, 0)
            iyhi = min(iy+t.ny, ny)
            jxlo = ixlo-ix
            jxhi = jxlo + max(0, ixhi - ixlo)
            jylo = iylo-iy
            jyhi = jylo + max(0, iyhi - iylo)
            if iyhi > iylo and ixhi > ixlo:
                data[iylo:iyhi, ixlo:ixhi] = t.data[jylo:jyhi, jxlo:jxhi]

    if not num.any(num.isfinite(data)):
        return None

    return Tile(xmin, ymin, dx, dy, data)


def tile_export(tle, path, format='stl', exaggeration=3., socket_scale=1.,
                scale=100):
    '''Export DEM to 3D printable formats'''
    import trimesh

    from pyrockoeost import geometry
    from pyrockoeost.cake import earthradius

    x = tle.x()
    x = x - x.min()

    y = tle.y()
    y = y - y.min()

    data = tle.data * exaggeration

    vertices, faces = geometry.topo_to_mesh(
        y, x, data, earthradius)

    vertices[:, 0] -= vertices[:, 0].min()

    nlat = y.size
    nlon = x.size
    socket_level = -vertices[:, 0].max() * socket_scale
    nvertices = vertices.shape[0]

    south_indices = num.arange(0, nlon)
    north_indices = num.arange(nvertices - nlon, nvertices)
    east_indices = num.arange(0, nlat) * nlon
    west_indices = num.arange(0, nlat) * nlon + nlon - 1

    north_indices = north_indices[::-1]
    east_indices = east_indices[::-1]

    socket_bottom_faces = []
    for border_indices in [north_indices, east_indices,
                           south_indices, west_indices]:
        nvertices = vertices.shape[0]
        nfaces = border_indices.size - 1

        socket_vertices = vertices[border_indices.astype(int)].copy()
        socket_vertices[:, 0] = socket_level

        socket_faces = num.empty((nfaces, 4), dtype=int)
        for iface in range(nfaces):
            socket_faces[iface, 0] = iface + nvertices
            socket_faces[iface, 1] = iface + nvertices + 1
            socket_faces[iface, 2] = border_indices[iface+1]
            socket_faces[iface, 3] = border_indices[iface]

        vertices = num.vstack((vertices, socket_vertices))
        faces = num.vstack((faces, socket_faces))

        socket_bottom_faces.append(vertices.shape[0] - nfaces)

    faces = num.vstack((faces, socket_bottom_faces[::-1]))

    mesh = trimesh.Trimesh(
        vertices, faces)
    mesh.fix_normals()
    mesh.fill_holes()

    mesh.apply_scale(scale)

    with open(path, 'wb') as f:
        mesh.export(f, file_type=format)

    return mesh



from __future__ import print_function, absolute_import

import sys
import logging
from optparse import OptionParser
from io import StringIO

import pyrockoeost
from pyrockoeost import deps
from pyrockoeost import util


logger = logging.getLogger('pyrockoeost.gui.drum.cli')
km = 1e3


def d2u(d):
    if isinstance(d, dict):
        return dict((k.replace('-', '_'), v) for (k, v) in d.items())
    else:
        return d.replace('-', '_')


subcommand_descriptions = {
    'view': 'open viewer (default)',
    'version': 'print version number',
}

subcommand_usages = {
    'view': 'view [options] <files> ...',
    'version': 'version',
}

subcommands = list(subcommand_descriptions.keys())

program_name = 'drum'

usage_tdata = d2u(subcommand_descriptions)
usage_tdata['program_name'] = program_name
usage_tdata['version_number'] = pyrockoeost.__version__


usage = '''%(program_name)s <subcommand> [options] [--] <arguments> ...

A helicorder and datalogger.

The Drum is part of pyrockoeost version %(version_number)s.

Subcommands:

    view            %(view)s
    version         %(version)s

To get further help and a list of available options for any subcommand run:

    %(program_name)s <subcommand> --help

''' % usage_tdata


def main(args=None):
    if not args:
        args = sys.argv

    help_triggers = ['--help', '-h', 'help']

    args = list(args)
    if len(args) < 2 or args[1] not in subcommands + help_triggers:
        args[1:1] = ['view']

    args.pop(0)
    command = args.pop(0)

    if command in subcommands:
        globals()['command_' + d2u(command)](args)

    elif command in help_triggers:
        if command == 'help' and args:
            acommand = args[0]
            if acommand in subcommands:
                globals()['command_' + acommand](['--help'])

        sys.exit('Usage: %s' % usage)

    else:
        die('No such subcommand: %s' % command)


def add_common_options(parser):
    parser.add_option(
        '--loglevel',
        action='store',
        dest='loglevel',
        type='choice',
        choices=('critical', 'error', 'warning', 'info', 'debug'),
        default='info',
        help='set logger level to '
             '"critical", "error", "warning", "info", or "debug". '
             'Default is "%default".')


def process_common_options(command, parser, options):
    util.setup_logging(program_name, options.loglevel)


def cl_parse(command, args, setup=None, details=None):
    usage = subcommand_usages[command]
    descr = subcommand_descriptions[command]

    if isinstance(usage, str):
        usage = [usage]

    susage = '%s %s' % (program_name, usage[0])
    for s in usage[1:]:
        susage += '\n%s%s %s' % (' '*7, program_name, s)

    description = descr[0].upper() + descr[1:] + '.'

    if details:
        description = description + '\n\n%s' % details

    parser = OptionParser(usage=susage, description=description)

    if setup:
        setup(parser)

    add_common_options(parser)
    (options, args) = parser.parse_args(args)
    process_common_options(command, parser, options)
    return parser, options, args


def die(message, err='', prelude=''):
    if prelude:
        prelude = prelude + '\n'

    if err:
        err = '\n' + err

    sys.exit('%s%s failed: %s%s' % (prelude, program_name, message, err))


def help_and_die(parser, message):
    sio = StringIO()
    parser.print_help(sio)
    die(message, prelude=sio.getvalue())


def command_view(args):
    def setup(parser):
        pass

    parser, options, args = cl_parse('view', args, setup)

    try:
        pyrockoeost.drum()
    except deps.MissingpyrockoeostDependency as e:
        die(str(e))


def command_version(args):
    def setup(parser):
        parser.add_option(
            '--short', dest='short', action='store_true',
            help='only print pyrockoeost\'s version number')

    parser, options, args = cl_parse('version', args, setup)

    from pyrockoeost.print_version import print_version
    print_version(not options.short)


if __name__ == '__main__':
    main()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import math
import re

from pyrockoeost import util


g_tmin, g_tmax = util.get_working_system_time_range()[:2]


def time_or_none_to_str(x, format):
    if x is None:
        return '...'
    else:
        return util.time_to_str(x, format=format)


def shorten_time_str(s):
    return re.sub(r'(((((-01)?-01)? 00)?:00)?:00)?(\.0+)?$', '', s)


def squirrel_content(cls):

    def str_codes(self):
        return str(self.codes)

    cls.str_codes = property(str_codes)

    def str_time_span(self):
        tmin, tmax = self.time_span
        deltat = getattr(self, 'deltat', 0)
        if deltat is not None and deltat > 0:
            fmt = min(9, max(0, -int(math.floor(math.log10(self.deltat)))))
        else:
            fmt = 6

        if tmin == tmax:
            return '%s' % time_or_none_to_str(tmin, fmt)
        else:
            return '%s - %s' % (
                time_or_none_to_str(tmin, fmt), time_or_none_to_str(tmax, fmt))

    cls.str_time_span = property(str_time_span)

    def str_time_span_short(self):
        return ' - '.join(
            shorten_time_str(x) for x in self.str_time_span.split(' - '))

    cls.str_time_span_short = property(str_time_span_short)

    def summary_entries(self):
        return (
            self.__class__.__name__,
            str(self.codes),
            self.str_time_span)

    if not hasattr(cls, 'summary_entries'):
        cls.summary_entries = property(summary_entries)

    def summary(self):
        return util.fmt_summary(self.summary_entries, (10, 20, 0))

    if not hasattr(cls, 'summary'):
        cls.summary = property(summary)

    def __lt__(self, other):
        return self.__key__() < other.__key__()

    cls.__lt__ = __lt__

    def __key__(self):
        return self.codes, self.time_span_g_clipped

    cls.__key__ = __key__

    @property
    def time_span_g_clipped(self):
        tmin, tmax = self.time_span
        return (
            tmin if tmin is not None else g_tmin,
            tmax if tmax is not None else g_tmax)

    cls.time_span_g_clipped = property(time_span_g_clipped)

    return cls



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import difflib
from collections import defaultdict

from pyrockoeost import util
from pyrockoeost.guts import Object, List, clone, path_to_str
from weakref import ref


class listdict(dict):
    def __missing__(self, k):
        self[k] = []
        return self[k]


def root_and_path(obj, name=None):
    root = obj
    path = []
    if name is not None:
        path.append(name)

    while True:
        try:
            root, name_at_parent = root._talkie_parent
            path.append(name_at_parent)
        except AttributeError:
            break

    return root, '.'.join(path[::-1])


def lclone(xs):
    return [clone(x) for x in xs]


g_uid = 0


def new_uid():
    global g_uid
    g_uid += 1
    return '#%i' % g_uid


def has_computed(cls):
    cls._computed = defaultdict(list)
    cls._computed_rev = defaultdict(list)

    for prop_name in dir(cls):
        try:
            prop = getattr(cls, prop_name)
            depends_on = prop.fget._computed_depends_on
            for name in depends_on:
                cls._computed_rev[name].append(prop_name)
                cls._computed[prop_name].append(name)

        except AttributeError:
            pass

    return cls


def computed(depends_on):
    def wrapper(func):
        func._computed_depends_on = depends_on
        return property(func)

    return wrapper


class Talkie(Object):
    _computed = ()
    _computed_rev = None

    def __setattr__(self, name, value):
        try:
            t = self.T.get_property(name)
        except ValueError:
            Object.__setattr__(self, name, value)
            return

        if isinstance(t, List.T):
            value = TalkieList(value)

        oldvalue = getattr(self, name, None)
        if oldvalue:
            if isinstance(oldvalue, (Talkie, TalkieList)):
                oldvalue.unset_parent()

        if isinstance(value, (Talkie, TalkieList)):
            value.set_parent(self, name)

        Object.__setattr__(self, name, value)
        self.fire([name], value)
        if self._computed_rev is not None:
            for dname in self._computed_rev[name]:
                self.fire([dname], None)

    def fire(self, path, value):
        self.fire_event(path, value)
        if hasattr(self, '_talkie_parent'):
            root, name_at_parent = self._talkie_parent
            path.append(name_at_parent)
            root.fire(path, value)

    def set_parent(self, parent, name):
        Object.__setattr__(self, '_talkie_parent', (parent, name))

    def unset_parent(self):
        Object.__delattr__(self, '_talkie_parent')

    def fire_event(self, path, value):
        pass

    def diff(self, other, path=()):
        assert type(self) is type(other), '%s %s' % (type(self), type(other))

        for (s_prop, s_val), (o_prop, o_val) in zip(
                self.T.ipropvals(self), other.T.ipropvals(other)):

            if not s_prop.multivalued:
                if isinstance(s_val, Talkie) \
                        and type(s_val) is type(o_val):

                    for x in s_val.diff(o_val, path + (s_prop.name,)):
                        yield x
                else:
                    if not equal(s_val, o_val):
                        yield 'set', path + (s_prop.name,), clone(o_val)
            else:
                if issubclass(s_prop.content_t.cls, Talkie):
                    sm = difflib.SequenceMatcher(
                        None,
                        type_eq_proxy_seq(s_val),
                        type_eq_proxy_seq(o_val))
                    mode = 1

                else:
                    sm = difflib.SequenceMatcher(
                        None,
                        eq_proxy_seq(s_val),
                        eq_proxy_seq(o_val))
                    mode = 2

                for tag, i1, i2, j1, j2 in list(sm.get_opcodes()):
                    if tag == 'equal' and mode == 1:
                        for koff, (s_element, o_element) in enumerate(zip(
                                s_val[i1:i2], o_val[j1:j2])):

                            for x in s_element.diff(
                                    o_element,
                                    path + ((s_prop.name, i1+koff),)):
                                yield x

                    if tag == 'replace':
                        yield (
                            'replace',
                            path + ((s_prop.name, i1, i2),),
                            lclone(o_val[j1:j2]))

                    elif tag == 'delete':
                        yield (
                            'delete',
                            path + ((s_prop.name, i1, i2),),
                            None)

                    elif tag == 'insert':
                        yield (
                            'insert',
                            path + ((s_prop.name, i1, i2),),
                            lclone(o_val[j1:j2]))

    def str_diff(self, other):
        lines = []
        for tag, path, value in self.diff(other):
            lines.append('%s %s' % (tag, path_to_str(path)))

        return '\n'.join(lines)

    def diff_update(self, other, path=()):
        assert type(self) is type(other), '%s %s' % (type(self), type(other))

        for (s_prop, s_val), (o_prop, o_val) in zip(
                self.T.ipropvals(self), other.T.ipropvals(other)):

            if not s_prop.multivalued:
                if isinstance(s_val, Talkie) \
                        and type(s_val) is type(o_val):

                    s_val.diff_update(o_val, path + (s_prop.name,))
                else:
                    if not equal(s_val, o_val):
                        setattr(self, s_prop.name, clone(o_val))
            else:
                if issubclass(s_prop.content_t.cls, Talkie):
                    sm = difflib.SequenceMatcher(
                        None,
                        type_eq_proxy_seq(s_val),
                        type_eq_proxy_seq(o_val))
                    mode = 1

                else:
                    sm = difflib.SequenceMatcher(
                        None,
                        eq_proxy_seq(s_val),
                        eq_proxy_seq(o_val))
                    mode = 2

                ioff = 0
                for tag, i1, i2, j1, j2 in list(sm.get_opcodes()):
                    if tag == 'equal' and mode == 1:
                        for koff, (s_element, o_element) in enumerate(zip(
                                s_val[i1+ioff:i2+ioff], o_val[j1:j2])):

                            s_element.diff_update(
                                o_element,
                                path + ((s_prop.name, i1+ioff+koff),))

                    elif tag == 'replace':
                        for _ in range(i1, i2):
                            s_val.pop(i1+ioff)

                        for j in range(j1, j2):
                            s_val.insert(i1+ioff, clone(o_val[j]))

                        ioff += (j2 - j1) - (i2 - i1)

                    elif tag == 'delete':
                        for _ in range(i1, i2):
                            s_val.pop(i1 + ioff)

                        ioff -= (i2 - i1)

                    elif tag == 'insert':
                        for j in range(j1, j2):
                            s_val.insert(i1+ioff, clone(o_val[j]))

                        ioff += (j2 - j1)


def equal(a, b):
    return str(a) == str(b)  # to be replaced by recursive guts.equal


def ghash(obj):
    return hash(str(obj))


class GutsEqProxy(object):
    def __init__(self, obj):
        self._obj = obj

    def __eq__(self, other):
        return equal(self._obj, other._obj)

    def __hash__(self):
        return ghash(self._obj)


class TypeEqProxy(object):
    def __init__(self, obj):
        self._obj = obj

    def __eq__(self, other):
        return type(self._obj) is type(other._obj)  # noqa

    def __hash__(self):
        return hash(type(self._obj))


def eq_proxy_seq(seq):
    return list(GutsEqProxy(x) for x in seq)


def type_eq_proxy_seq(seq):
    return list(TypeEqProxy(x) for x in seq)


class TalkieConnection(object):
    def __init__(self, talkie_root, path, listener):
        self._talkie_root = talkie_root
        self._listener = listener
        self._path = path
        self._ref_listener = ref(listener)

    def release(self):
        self._talkie_root.disconnect(self)


class TalkieRoot(Talkie):

    def __init__(self, **kwargs):
        self._listeners = listdict()
        Talkie.__init__(self, **kwargs)

    def talkie_connect(self, path, listener):
        connection = TalkieConnection(self, path, listener)
        self._listeners[path].append(connection._ref_listener)
        return connection

    def talkie_disconnect(self, connection):
        try:
            self._listeners[connection._path].remove(
                connection._ref_listener)
        except ValueError:
            pass

    def fire_event(self, path, value):
        path = '.'.join(path[::-1])
        # print('fire_event:', path, value)
        parts = path.split('.')
        for i in range(len(parts)+1):
            subpath = '.'.join(parts[:i])
            target_refs = self._listeners[subpath]
            delete = []
            for target_ref in target_refs:
                target = target_ref()
                if target:
                    target(path, value)
                else:
                    delete.append(target_ref)

            for target_ref in delete:
                target_refs.remove(target_ref)

    def get(self, path):
        x = self
        for s in path.split('.'):
            x = getattr(x, s)

        return x

    def set(self, path, value):
        x = self
        p = path.split('.')
        for s in p[:-1]:
            x = getattr(x, s)

        setattr(x, p[-1], value)


class TalkieList(list):

    def fire(self, path, value):
        if self._talkie_parent:
            root, name_at_parent = self._talkie_parent
            path.append(name_at_parent)
            root.fire(path, value)

    def set_parent(self, parent, name):
        list.__setattr__(self, '_talkie_parent', (parent, name))

    def unset_parent(self):
        list.__delattr__(self, '_talkie_parent')

    def append(self, element):
        retval = list.append(self, element)
        name = new_uid()
        if isinstance(element, (Talkie, TalkieList)):
            element.set_parent(self, name)

        self.fire([], self)
        return retval

    def insert(self, index, element):
        retval = list.insert(self, index, element)
        name = new_uid()
        if isinstance(element, (Talkie, TalkieList)):
            element.set_parent(self, name)

        self.fire([], self)
        return retval

    def remove(self, element):
        list.remove(self, element)
        if isinstance(element, (Talkie, TalkieList)):
            element.unset_parent()

        self.fire([], self)

    def pop(self, index=-1):
        element = list.pop(self, index)
        if isinstance(element, (Talkie, TalkieList)):
            element.unset_parent()

        self.fire([], self)
        return element

    def extend(self, elements):
        for element in elements:
            self.append(element)

        self.fire([], self)

    def __setitem__(self, key, value):
        try:
            element = self[key]
            if isinstance(element, (Talkie, TalkieList)):
                element.unset_parent()

        except IndexError:
            pass

        list.__setitem__(self, key, value)
        self.fire([], self)

    def __setslice__(self, *args, **kwargs):
        raise Exception('not implemented')

    def __iadd__(self, *args, **kwargs):
        raise Exception('not implemented')

    def __imul__(self, *args, **kwargs):
        raise Exception('not implemented')


for method_name in ['reverse', 'sort']:

    def x():
        list_meth = getattr(list, method_name)

        def meth(self, *args, **kwargs):
            retval = list_meth(self, *args, **kwargs)
            self.fire([], self)
            return retval

        return meth

    try:
        setattr(TalkieList, method_name, x())
    except AttributeError:
        pass


def drop_args_wrapper(f):
    def f_drop_args(*args):
        f()

    return f_drop_args


class TalkieConnectionOwner(object):
    def __init__(self):
        self._connections = []

    def talkie_connect(self, state, path, listener, drop_args=False):
        if drop_args:
            listener = drop_args_wrapper(listener)

        if not isinstance(path, str):
            return [
                self.talkie_connect(state, path_, listener, False)
                for path_ in path]

        connection = state.talkie_connect(path, listener)
        self._connections.append(connection)
        return connection

    def talkie_disconnect_all(self):
        while self._connections:
            try:
                self._connections.pop().release()
            except Exception:
                pass


def none_or(f):
    def g(x):
        if x is None:
            return ''
        else:
            return f(x)

    return g


def noop(x):
    return x


class TalkieStringer:
    def __init__(self, root):
        self._root = root
        self._formatters = {
            'date': none_or(lambda v: util.time_to_str(v, format='%Y-%m-%d')),
            'datetime': none_or(lambda v: util.time_to_str(v))}

        self._paths = []

    def __getitem__(self, key):

        stringer = self

        class Proxy:
            def __init__(self, val, formatter, path, talkie_root):
                self._formatter = formatter
                self._val = val
                self._path = path
                self._talkie_root = talkie_root

            def register_path(self):
                stringer._paths.append(
                    (self._talkie_root, '.'.join(self._path)))

            def _computed(self):
                return getattr(self._val, '_computed', ())

            def __getattr__(self, key):
                if key in self._val.T.propnames \
                        or key in self._computed():

                    self._path.append(key)
                    val = getattr(self._val, key)
                    pval = self._proxy(val)
                    if not isinstance(pval, (Proxy, ListProxy)):
                        self.register_path()

                    return pval
                else:
                    return '{' + key + '}'

            def __format__(self, *args, **kwargs):
                self.register_path()
                return self._val.__format__(*args, **kwargs)

            def _proxy(self, val):
                if isinstance(val, TalkieRoot):
                    return Proxy(
                        val, self._formatter, [], val)

                elif isinstance(val, Object):
                    return Proxy(
                        val, self._formatter, self._path, self._talkie_root)

                elif isinstance(val, list):
                    return ListProxy(
                        val, self._formatter, self._path, self._talkie_root)

                else:
                    return self._formatter(val)

        class ListProxy(Proxy):
            def __getitem__(self, i):
                self._path[-1] += '[%i]' % i
                return self._proxy(self._val[i])

        key = key.split('|', 1)
        if len(key) == 2:
            key, formatter = key[0], self._formatters[key[1]]
        else:
            key, formatter = key[0], noop

        return getattr(Proxy(self._root, formatter, [], self._root), key)

    def get_paths(self):
        return self._paths



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import copy

import numpy as num

from pyrockoeost.guts import Bool, Float, String, StringChoice
from pyrockoeost.gui import util as gui_util
from pyrockoeost.gui.vtk_util import ScatterPipe, BeachballPipe
from pyrockoeost.gui.qt_compat import qw, qc

from . import base
from .. import common

guts_prefix = 'sparrow'
km = 1e3


def inormalize(x, imin, imax, discrete=True):
    if x.size == 0:
        return num.full(x.shape, imin, dtype=int)

    xmin = num.nanmin(x)
    xmax = num.nanmax(x)
    if xmin == xmax:
        xmin -= 0.5
        xmax += 0.5

    rmin = imin - 0.5
    rmax = imax + 0.5

    if discrete:
        return num.clip(
            num.round(
                rmin + (x - xmin) * (
                    (rmax-rmin) / (xmax - xmin))).astype(num.int64),
            imin, imax)
    else:
        return num.clip(
            rmin + (x - xmin) * ((rmax-rmin) / (xmax - xmin)),
            imin, imax)


def string_to_sorted_idx(values):
    val_sort = num.sort(values, axis=-1, kind='mergesort')
    val_sort_unique = num.unique(val_sort)

    val_to_idx = dict([
        (val_sort_unique[i], i)
        for i in range(val_sort_unique.shape[0])])

    return num.array([val_to_idx[val] for val in values])


class SymbolChoice(StringChoice):
    choices = ['point', 'sphere', 'beachball']


class MaskingShapeChoice(StringChoice):
    choices = ['rect', 'linear', 'quadratic']


class MaskingModeChoice(StringChoice):
    choices = ['past + future', 'past', 'future']

    @classmethod
    def get_factors(cls, mode, value_low):
        return {
            'past + future': (value_low, 1.0, value_low),
            'past': (value_low, 1.0, 0.0),
            'future': (0.0, 1.0, value_low)}[mode]


class TableState(base.ElementState):
    visible = Bool.T(default=True)
    size = Float.T(default=3.0)
    color_parameter = String.T(optional=True)
    cpt = base.CPTState.T(default=base.CPTState.D())
    size_parameter = String.T(optional=True)
    depth_min = Float.T(optional=True)
    depth_max = Float.T(optional=True)
    depth_offset = Float.T(default=0.0)
    symbol = SymbolChoice.T(default='sphere')
    time_masking_opacity = Float.T(default=0.0)
    time_masking_shape = MaskingShapeChoice.T(default='rect')
    time_masking_mode = MaskingModeChoice.T(default='past + future')


class TableElement(base.Element):
    def __init__(self):
        base.Element.__init__(self)
        self._parent = None

        self._table = None
        self._istate = 0
        self._istate_view = 0

        self._controls = None
        self._color_combobox = None
        self._size_combobox = None

        self._pipes = None
        self._pipe_maps = None
        self._isize_min = 1
        self._isize_max = 6

        self.cpt_handler = base.CPTHandler()

    def bind_state(self, state):
        base.Element.bind_state(self, state)
        self.talkie_connect(state, ['visible', 'size'], self.update)

        self.talkie_connect(
            state,
            ['depth_min', 'depth_max', 'time_masking_shape',
             'time_masking_mode', 'time_masking_opacity'],
            self.update_alpha)

        self.cpt_handler.bind_state(state.cpt, self.update)

        self.talkie_connect(
            state,
            ['symbol', 'size_parameter', 'color_parameter'],
            self.update_sizes)

    def unbind_state(self):
        self.cpt_handler.unbind_state()
        base.Element.unbind_state(self)

    def get_name(self):
        return 'Table'

    def set_parent(self, parent):
        self._parent = parent
        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        for var in ['tmin', 'tmax', 'tduration', 'tposition']:
            self.talkie_connect(
                self._parent.state, var, self.update_alpha)

        self._parent.register_data_provider(self)

        self.update()

    def iter_data(self, name):
        if self._table and self._table.has_col(name):
            yield self._table.get_col(name)

    def set_table(self, table):
        self._table = table

        self._istate += 1

        if self._pipes is not None and self._istate != self._istate_view:
            self._clear_pipes()

        self._update_controls()

    def get_size_parameter_extra_entries(self):
        return []

    def get_color_parameter_extra_entries(self):
        return []

    def update_sizes(self, *args):
        self._istate += 1
        self.update()

    def unset_parent(self):
        self.unbind_state()
        if self._parent:
            self._parent.unregister_data_provider(self)

            self._clear_pipes()

            if self._controls:
                self._parent.remove_panel(self._controls)
                self._controls = None

            self._parent.update_view()
            self._parent = None

    def _clear_pipes(self):
        if self._pipes is not None:
            for p in self._pipes:
                self._parent.remove_actor(p.actor)

            self._pipes = None

        if self._pipe_maps is not None:
            self._pipe_maps = None

    def _init_pipes_scatter(self):
        state = self._state
        points = self._table.get_col('xyz')
        self._pipes = []
        self._pipe_maps = []
        if state.size_parameter:
            sizes = self._table.get_col(state.size_parameter)
            isizes = inormalize(
                sizes, self._isize_min, self._isize_max)

            for i in range(self._isize_min, self._isize_max+1):
                b = isizes == i
                p = ScatterPipe(points[b].copy())
                self._pipes.append(p)
                self._pipe_maps.append(b)
        else:
            self._pipes.append(
                ScatterPipe(points))
            self._pipe_maps.append(
                num.ones(points.shape[0], dtype=bool))

    def _init_pipes_beachball(self):
        state = self._state
        self._pipes = []

        tab = self._table

        positions = tab.get_col('xyz')

        if tab.has_col('m6'):
            m6s = tab.get_col('m6')
        else:
            m6s = num.zeros((tab.get_nrows(), 6))
            m6s[:, 3] = 1.0

        if state.size_parameter:
            sizes = tab.get_col(state.size_parameter)
        else:
            sizes = num.ones(tab.get_nrows())

        if state.color_parameter:
            values = self._table.get_col(state.color_parameter)
        else:
            values = num.zeros(tab.get_nrows())

        rsizes = inormalize(
            sizes, self._isize_min, self._isize_max, discrete=False) * 0.005

        pipe = BeachballPipe(positions, m6s, rsizes, values, self._parent.ren)
        self._pipes = [pipe]

    def _update_pipes_scatter(self):
        state = self._state
        for i, p in enumerate(self._pipes):
            self._parent.add_actor(p.actor)
            p.set_size(state.size * (self._isize_min + i)**1.3)

        if state.color_parameter:
            values = self._table.get_col(state.color_parameter)

            if num.issubdtype(values.dtype, num.string_):
                values = string_to_sorted_idx(values)

            self.cpt_handler._values = values
            self.cpt_handler.update_cpt()

            cpt = copy.deepcopy(
                self.cpt_handler._cpts[self._state.cpt.effective_cpt_name])
            colors2 = cpt(values)
            colors2 = colors2 / 255.

            for m, p in zip(self._pipe_maps, self._pipes):
                p.set_colors(colors2[m, :])

        for p in self._pipes:
            p.set_symbol(state.symbol)

    def _update_pipes_beachball(self):
        state = self._state

        p = self._pipes[0]

        self._parent.add_actor(p.actor)
        p.set_size_factor(state.size * 0.005)

    def _init_pipes(self):
        if self._state.symbol == 'beachball':
            self._init_pipes_beachball()
        else:
            self._init_pipes_scatter()

    def _update_pipes(self):
        if self._state.symbol == 'beachball':
            self._update_pipes_beachball()
        else:
            self._update_pipes_scatter()

    def update(self, *args):
        state = self._state

        if self._pipes is not None and self._istate != self._istate_view:
            self._clear_pipes()

        if not state.visible:
            if self._pipes is not None:
                for p in self._pipes:
                    self._parent.remove_actor(p.actor)

        else:
            if self._istate != self._istate_view and self._table:
                self._init_pipes()
                self._istate_view = self._istate

            if self._pipes is not None:
                self._update_pipes()

        self.update_alpha()  # TODO: only if needed?
        self._parent.update_view()

    def update_alpha(self, *args, mask=None):
        if self._state.symbol == 'beachball':
            return

        if self._pipes is None:
            return

        time = self._table.get_col('time')
        depth = self._table.get_col('depth')

        depth_mask = num.ones(time.size, dtype=bool)

        if self._state.depth_min is not None:
            depth_mask &= depth >= self._state.depth_min
        if self._state.depth_max is not None:
            depth_mask &= depth <= self._state.depth_max

        tmin = self._parent.state.tmin_effective
        tmax = self._parent.state.tmax_effective

        if tmin is not None:
            m1 = time < tmin
        else:
            m1 = num.zeros(time.size, dtype=bool)

        if tmax is not None:
            m3 = tmax < time
        else:
            m3 = num.zeros(time.size, dtype=bool)

        m2 = num.logical_not(num.logical_or(m1, m3))

        value_low = self._state.time_masking_opacity

        f1, f2, f3 = MaskingModeChoice.get_factors(
            self._state.time_masking_mode, value_low)

        amp = num.ones(time.size, dtype=num.float64)
        amp[m1] = f1
        amp[m3] = f3
        if None in (tmin, tmax):
            amp[m2] = 1.0
        else:
            if self._state.time_masking_shape == 'rect':
                amp[m2] == 1.0
            elif self._state.time_masking_shape == 'linear':
                amp[m2] = time[m2]
                amp[m2] -= tmin
                amp[m2] /= (tmax - tmin)
            elif self._state.time_masking_shape == 'quadratic':
                amp[m2] = time[m2]
                amp[m2] -= tmin
                amp[m2] /= (tmax - tmin)
                amp[m2] **= 2

            if f1 != 0.0:
                amp[m2] *= (1.0 - value_low)
                amp[m2] += value_low

        amp *= depth_mask

        for m, p in zip(self._pipe_maps, self._pipes):
            p.set_alpha(amp[m])

        self._parent.update_view()

    def _get_table_widgets_start(self):
        return 0

    def _get_controls(self):
        if self._controls is None:
            from ..state import state_bind_slider, state_bind_slider_float, \
                state_bind_combobox, state_bind_lineedit

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            frame.setLayout(layout)

            iy = self._get_table_widgets_start()

            layout.addWidget(qw.QLabel('Size'), iy, 0)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(100)
            layout.addWidget(slider, iy, 1)
            state_bind_slider(self, self._state, 'size', slider, factor=0.1)

            iy += 1

            layout.addWidget(qw.QLabel('Size Scaling'), iy, 0)

            cb = qw.QComboBox()

            layout.addWidget(cb, iy, 1)
            state_bind_combobox(
                self, self._state, 'size_parameter', cb)

            self._size_combobox = cb

            iy += 1

            layout.addWidget(qw.QLabel('Color'), iy, 0)

            cb = qw.QComboBox()

            layout.addWidget(cb, iy, 1)
            state_bind_combobox(
                self, self._state, 'color_parameter', cb)

            self._color_combobox = cb

            self.cpt_handler.cpt_controls(
                self._parent, self._state.cpt, layout)

            iy = layout.rowCount() + 1

            layout.addWidget(qw.QLabel('Symbol'), iy, 0)

            cb = common.string_choices_to_combobox(SymbolChoice)

            layout.addWidget(cb, iy, 1)
            state_bind_combobox(
                self, self._state, 'symbol', cb)

            iy += 1

            layout.addWidget(qw.QLabel('Depth Min [km]'), iy, 0)
            slider = gui_util.QSliderFloat(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimumFloat(-60*km)
            slider.setMaximumFloat(700*km)
            layout.addWidget(slider, iy, 1)
            state_bind_slider_float(
                self, self._state, 'depth_min', slider,
                min_is_none=True)
            self._depth_min_slider = slider

            le = qw.QLineEdit()
            layout.addWidget(le, iy, 2)
            state_bind_lineedit(
                self, self._state, 'depth_min', le,
                from_string=lambda s: None if s == 'off' else float(s)*1000.,
                to_string=lambda v: 'off' if v is None else str(v/1000.))

            self._depth_min_lineedit = le

            iy += 1

            layout.addWidget(qw.QLabel('Depth Max [km]'), iy, 0)
            slider = gui_util.QSliderFloat(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimumFloat(-60*km)
            slider.setMaximumFloat(700*km)
            layout.addWidget(slider, iy, 1)
            state_bind_slider_float(
                self, self._state, 'depth_max', slider,
                max_is_none=True)
            self._depth_max_slider = slider

            le = qw.QLineEdit()
            layout.addWidget(le, iy, 2)
            state_bind_lineedit(
                self, self._state, 'depth_max', le,
                from_string=lambda s: None if s == 'off' else float(s)*1000.,
                to_string=lambda v: 'off' if v is None else str(v/1000.))

            self._depth_max_lineedit = le

            iy += 1

            layout.addWidget(qw.QLabel('Time Masking Opacity'), iy, 0)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(100)
            layout.addWidget(slider, iy, 1)
            state_bind_slider(
                self, self._state, 'time_masking_opacity', slider, factor=0.01)

            iy += 1

            layout.addWidget(qw.QLabel('Time Masking Shape'), iy, 0)
            cb = common.string_choices_to_combobox(MaskingShapeChoice)
            layout.addWidget(cb, iy, 1)
            state_bind_combobox(self, self._state, 'time_masking_shape', cb)

            iy += 1

            layout.addWidget(qw.QLabel('Time Masking Mode'), iy, 0)
            cb = common.string_choices_to_combobox(MaskingModeChoice)
            layout.addWidget(cb, iy, 1)
            state_bind_combobox(self, self._state, 'time_masking_mode', cb)

            iy += 1

            layout.addWidget(qw.QFrame(), iy, 0, 1, 3)

            self._controls = frame

            self._update_controls()

        return self._controls

    def _update_controls(self):
        for (cb, get_extra_entries) in [
                (self._color_combobox, self.get_color_parameter_extra_entries),
                (self._size_combobox, self.get_size_parameter_extra_entries)]:

            if cb is not None:
                cb.clear()

                have = set()
                for s in get_extra_entries():
                    if s not in have:
                        cb.insertItem(len(have), s)
                        have.add(s)

                if self._table is not None:
                    for s in self._table.get_col_names():
                        h = self._table.get_header(s)
                        if h.get_ncols() == 1 and s not in have:
                            cb.insertItem(len(have), s)
                            have.add(s)

        self.cpt_handler._update_cpt_combobox()
        self.cpt_handler._update_cptscale_lineedit()

        if self._table is not None and self._table.has_col('depth'):
            depth = self._table.get_col('depth')

            if depth.size > 0:

                depth_min = depth.min()
                depth_max = depth.max()

                for wdg in (self._depth_min_slider, self._depth_max_slider):
                    wdg.setMinimumFloat(depth_min)
                    wdg.setMaximumFloat(depth_max)


__all__ = [
    'TableElement',
    'TableState',
]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import re
import logging
import os.path as op

import numpy as num

from pyrockoeost import io, trace, util
from pyrockoeost.progress import progress
from pyrockoeost.has_paths import Path, HasPaths
from pyrockoeost.guts import Dict, String, Choice, Float, List, Timestamp, \
    StringChoice, IntChoice, Defer, load_all, clone
from pyrockoeost.squirrel.dataset import Dataset
from pyrockoeost.squirrel.client.local import LocalData
from pyrockoeost.squirrel.error import ToolError
from pyrockoeost.squirrel.model import CodesNSLCE
from pyrockoeost.squirrel.operators.base import NetworkGrouping, StationGrouping, \
    ChannelGrouping, SensorGrouping

tts = util.time_to_str

guts_prefix = 'jackseis'
logger = logging.getLogger('psq.cli.jackseis')


def make_task(*args):
    return progress.task(*args, logger=logger)


def parse_rename_rule_from_string(s):
    s = s.strip()
    if re.match(r'^([^:,]*:[^:,]*,?)+', s):
        return dict(
            x.split(':') for x in s.strip(',').split(','))
    else:
        return s


class JackseisError(ToolError):
    pass


class Chain(object):
    def __init__(self, node, parent=None):
        self.node = node
        self.parent = parent

    def mcall(self, name, *args, **kwargs):
        ret = []
        if self.parent is not None:
            ret.append(self.parent.mcall(name, *args, **kwargs))

        ret.append(getattr(self.node, name)(*args, **kwargs))
        return ret

    def fcall(self, name, *args, **kwargs):
        v = getattr(self.node, name)(*args, **kwargs)
        if v is None and self.parent is not None:
            return self.parent.fcall(name, *args, **kwargs)
        else:
            return v

    def get(self, name):
        v = getattr(self.node, name)
        if v is None and self.parent is not None:
            return self.parent.get(name)
        else:
            return v

    def dget(self, name, k):
        v = getattr(self.node, name).get(k, None)
        if v is None and self.parent is not None:
            return self.parent.dget(name, k)
        else:
            return v


class OutputFormatChoice(StringChoice):
    choices = io.allowed_formats('save')


class OutputDataTypeChoice(StringChoice):
    choices = ['int32', 'int64', 'float32', 'float64']
    name_to_dtype = {
        'int32': num.int32,
        'int64': num.int64,
        'float32': num.float32,
        'float64': num.float64}


class TraversalChoice(StringChoice):
    choices = ['network', 'station', 'channel', 'sensor']
    name_to_grouping = {
        'network': NetworkGrouping(),
        'station': StationGrouping(),
        'sensor': SensorGrouping(),
        'channel': ChannelGrouping()}


class Converter(HasPaths):

    in_dataset = Dataset.T(optional=True)
    in_path = String.T(optional=True)
    in_paths = List.T(String.T(optional=True))

    codes = List.T(CodesNSLCE.T(), optional=True)

    rename = Dict.T(
        String.T(),
        Choice.T([
            String.T(),
            Dict.T(String.T(), String.T())]))
    tmin = Timestamp.T(optional=True)
    tmax = Timestamp.T(optional=True)
    tinc = Float.T(optional=True)

    downsample = Float.T(optional=True)

    out_path = Path.T(optional=True)
    out_sds_path = Path.T(optional=True)
    out_format = OutputFormatChoice.T(optional=True)
    out_data_type = OutputDataTypeChoice.T(optional=True)
    out_mseed_record_length = IntChoice.T(
        optional=True,
        choices=list(io.mseed.VALID_RECORD_LENGTHS))
    out_mseed_steim = IntChoice.T(
        optional=True,
        choices=[1, 2])
    out_meta_path = Path.T(optional=True)

    traversal = TraversalChoice.T(optional=True)

    parts = List.T(Defer('Converter.T'))

    @classmethod
    def add_arguments(cls, p):
        p.add_squirrel_query_arguments(without=['time'])

        p.add_argument(
            '--tinc',
            dest='tinc',
            type=float,
            metavar='SECONDS',
            help='Set time length of output files [s].')

        p.add_argument(
            '--downsample',
            dest='downsample',
            type=float,
            metavar='RATE',
            help='Downsample to RATE [Hz].')

        p.add_argument(
            '--out-path',
            dest='out_path',
            metavar='TEMPLATE',
            help='Set output path to TEMPLATE. Available placeholders '
                 'are %%n: network, %%s: station, %%l: location, %%c: '
                 'channel, %%b: begin time, %%e: end time, %%j: julian day of '
                 'year. The following additional placeholders use the window '
                 'begin and end times rather than trace begin and end times '
                 '(to suppress producing many small files for gappy traces), '
                 '%%(wmin_year)s, %%(wmin_month)s, %%(wmin_day)s, %%(wmin)s, '
                 '%%(wmin_jday)s, %%(wmax_year)s, %%(wmax_month)s, '
                 '%%(wmax_day)s, %%(wmax)s, %%(wmax_jday)s. '
                 "Example: --out-path='data/%%s/trace-%%s-%%c.mseed'")

        p.add_argument(
            '--out-sds-path',
            dest='out_sds_path',
            metavar='PATH',
            help='Set output path to create SDS (https://www.seiscomp.de'
                 '/seiscomp3/doc/applications/slarchive/SDS.html), rooted at '
                 'the given path. Implies --tinc=86400. '
                 'Example: --out-sds-path=data/sds')

        p.add_argument(
            '--out-format',
            dest='out_format',
            choices=io.allowed_formats('save'),
            metavar='FORMAT',
            help='Set output file format. Choices: %s' % io.allowed_formats(
                'save', 'cli_help', 'mseed'))

        p.add_argument(
            '--out-data-type',
            dest='out_data_type',
            choices=OutputDataTypeChoice.choices,
            metavar='DTYPE',
            help='Set numerical data type. Choices: %s. The output file '
                 'format must support the given type. By default, the data '
                 'type is kept unchanged.' % ', '.join(
                    OutputDataTypeChoice.choices))

        p.add_argument(
            '--out-mseed-record-length',
            dest='out_mseed_record_length',
            type=int,
            choices=io.mseed.VALID_RECORD_LENGTHS,
            metavar='INT',
            help='Set the mseed record length in bytes. Choices: %s. '
                 'Default is 4096 bytes, which is commonly used for archiving.'
                 % ', '.join(str(b) for b in io.mseed.VALID_RECORD_LENGTHS))

        p.add_argument(
            '--out-mseed-steim',
            dest='out_mseed_steim',
            type=int,
            choices=(1, 2),
            metavar='INT',
            help='Set the mseed STEIM compression. Choices: 1 or 2. '
                 'Default is STEIM-2, which can compress full range int32. '
                 'Note: STEIM-2 is limited to 30 bit dynamic range.')

        p.add_argument(
            '--out-meta-path',
            dest='out_meta_path',
            metavar='PATH',
            help='Set output path for station metadata (StationXML) export.')

        p.add_argument(
            '--traversal',
            dest='traversal',
            metavar='GROUPING',
            choices=TraversalChoice.choices,
            help='By default the outermost processing loop is over time. '
                 'Add outer loop with given GROUPING. Choices: %s'
                 % ', '.join(TraversalChoice.choices))

        p.add_argument(
            '--rename-network',
            dest='rename_network',
            metavar='REPLACEMENT',
            help="""
Replace network code. REPLACEMENT can be a string for direct replacement, a
mapping for selective replacement, or a regular expression for tricky
replacements. Examples: Direct replacement: ```XX``` - set all network codes to
```XX```. Mapping: ```AA:XX,BB:YY``` - replace ```AA``` with ```XX``` and
```BB``` with ```YY```. Regular expression: ```/A(\\d)/X\\1/``` - replace
```A1``` with ```X1``` and ```A2``` with ```X2``` etc.
""".strip())

        p.add_argument(
            '--rename-station',
            dest='rename_station',
            metavar='REPLACEMENT',
            help='Replace station code. See ``--rename-network``.')

        p.add_argument(
            '--rename-location',
            dest='rename_location',
            metavar='REPLACEMENT',
            help='Replace location code. See ``--rename-network``.')

        p.add_argument(
            '--rename-channel',
            dest='rename_channel',
            metavar='REPLACEMENT',
            help='Replace channel code. See ``--rename-network``.')

        p.add_argument(
            '--rename-extra',
            dest='rename_extra',
            metavar='REPLACEMENT',
            help='Replace extra code. See ``--rename-network``. Note: the '
                 '```extra``` code is not available in Mini-SEED.')

    @classmethod
    def from_arguments(cls, args):
        kwargs = args.squirrel_query

        rename = {}
        for (k, v) in [
                ('network', args.rename_network),
                ('station', args.rename_station),
                ('location', args.rename_location),
                ('channel', args.rename_channel),
                ('extra', args.rename_extra)]:

            if v is not None:
                rename[k] = parse_rename_rule_from_string(v)

        obj = cls(
            downsample=args.downsample,
            out_format=args.out_format,
            out_path=args.out_path,
            tinc=args.tinc,
            out_sds_path=args.out_sds_path,
            out_data_type=args.out_data_type,
            out_mseed_record_length=args.out_mseed_record_length,
            out_mseed_steim=args.out_mseed_steim,
            out_meta_path=args.out_meta_path,
            traversal=args.traversal,
            rename=rename,
            **kwargs)

        obj.validate()
        return obj

    def add_dataset(self, sq):
        if self.in_dataset is not None:
            sq.add_dataset(self.in_dataset)

        if self.in_path is not None:
            ds = Dataset(sources=[LocalData(paths=[self.in_path])])
            ds.set_basepath_from(self)
            sq.add_dataset(ds)

        if self.in_paths:
            ds = Dataset(sources=[LocalData(paths=self.in_paths)])
            ds.set_basepath_from(self)
            sq.add_dataset(ds)

    def get_effective_rename_rules(self, chain):
        d = {}
        for k in ['network', 'station', 'location', 'channel']:
            v = chain.dget('rename', k)
            if isinstance(v, str):
                m = re.match(r'/([^/]+)/([^/]*)/', v)
                if m:
                    try:
                        v = (re.compile(m.group(1)), m.group(2))
                    except Exception:
                        raise JackseisError(
                            'Invalid replacement pattern: /%s/' % m.group(1))

            d[k] = v

        return d

    def get_effective_out_path(self):
        nset = sum(x is not None for x in (
            self.out_path,
            self.out_sds_path))

        if nset > 1:
            raise JackseisError(
                'More than one out of [out_path, out_sds_path] set.')

        is_sds = False
        if self.out_path:
            out_path = self.out_path

        elif self.out_sds_path:
            out_path = op.join(
                self.out_sds_path,
                '%(wmin_year)s/%(network)s/%(station)s/%(channel)s.D'
                '/%(network)s.%(station)s.%(location)s.%(channel)s.D'
                '.%(wmin_year)s.%(wmin_jday)s')
            is_sds = True
        else:
            out_path = None

        if out_path is not None:
            return self.expand_path(out_path), is_sds
        else:
            return None

    def get_effective_out_meta_path(self):
        if self.out_meta_path is not None:
            return self.expand_path(self.out_meta_path)
        else:
            return None

    def do_rename(self, rules, tr):
        rename = {}
        for k in ['network', 'station', 'location', 'channel']:
            v = rules.get(k, None)
            if isinstance(v, str):
                rename[k] = v
            elif isinstance(v, dict):
                try:
                    oldval = getattr(tr, k)
                    rename[k] = v[oldval]
                except KeyError:
                    raise ToolError(
                        'No mapping defined for %s code "%s".' % (k, oldval))

            elif isinstance(v, tuple):
                pat, repl = v
                oldval = getattr(tr, k)
                newval, n = pat.subn(repl, oldval)
                if n:
                    rename[k] = newval

        tr.set_codes(**rename)

    def convert(self, args, chain=None):
        if chain is None:
            defaults = clone(g_defaults)
            defaults.set_basepath_from(self)
            chain = Chain(defaults)

        chain = Chain(self, chain)

        if self.parts:
            task = make_task('Jackseis parts')
            for part in task(self.parts):
                part.convert(args, chain)

            del task

        else:
            sq = args.make_squirrel()

            cli_overrides = Converter.from_arguments(args)
            cli_overrides.set_basepath('.')

            chain = Chain(cli_overrides, chain)

            chain.mcall('add_dataset', sq)

            tmin = chain.get('tmin')
            tmax = chain.get('tmax')
            tinc = chain.get('tinc')
            codes = chain.get('codes')
            downsample = chain.get('downsample')
            out_path, is_sds = chain.fcall('get_effective_out_path') \
                or (None, False)

            if is_sds and tinc != 86400.:
                logger.warning('Setting time window to 1 day to generate SDS.')
                tinc = 86400.0

            out_format = chain.get('out_format')
            out_data_type = chain.get('out_data_type')

            out_meta_path = chain.fcall('get_effective_out_meta_path')

            if out_meta_path is not None:
                sx = sq.get_stationxml(codes=codes, tmin=tmin, tmax=tmax)
                util.ensuredirs(out_meta_path)
                sx.dump_xml(filename=out_meta_path)
                if out_path is None:
                    return

            target_deltat = None
            if downsample is not None:
                target_deltat = 1.0 / float(downsample)

            save_kwargs = {}
            if out_format == 'mseed':
                save_kwargs['record_length'] = chain.get(
                    'out_mseed_record_length')
                save_kwargs['steim'] = chain.get(
                    'out_mseed_steim')

            traversal = chain.get('traversal')
            if traversal is not None:
                grouping = TraversalChoice.name_to_grouping[traversal]
            else:
                grouping = None

            tpad = 0.0
            if target_deltat is not None:
                tpad += target_deltat * 50.

            task = None
            rename_rules = self.get_effective_rename_rules(chain)
            for batch in sq.chopper_waveforms(
                    tmin=tmin, tmax=tmax, tpad=tpad, tinc=tinc,
                    codes=codes,
                    snap_window=True,
                    grouping=grouping):

                if task is None:
                    task = make_task(
                        'Jackseis blocks', batch.n * batch.ngroups)

                tlabel = '%s%s - %s' % (
                    'groups %i / %i: ' % (batch.igroup, batch.ngroups)
                    if batch.ngroups > 1 else '',
                    util.time_to_str(batch.tmin),
                    util.time_to_str(batch.tmax))

                task.update(batch.i + batch.igroup * batch.n, tlabel)

                twmin = batch.tmin
                twmax = batch.tmax

                traces = batch.traces

                if target_deltat is not None:
                    out_traces = []
                    for tr in traces:
                        try:
                            tr.downsample_to(
                                target_deltat, snap=True, demean=False,
                                allow_upsample_max=4)

                            out_traces.append(tr)

                        except (trace.TraceTooShort, trace.NoData):
                            pass

                    traces = out_traces

                for tr in traces:
                    self.do_rename(rename_rules, tr)

                if out_data_type:
                    for tr in traces:
                        tr.ydata = tr.ydata.astype(
                            OutputDataTypeChoice.name_to_dtype[out_data_type])

                chopped_traces = []
                for tr in traces:
                    try:
                        otr = tr.chop(twmin, twmax, inplace=False)
                        chopped_traces.append(otr)
                    except trace.NoData:
                        pass

                traces = chopped_traces

                if out_path is not None:
                    try:
                        io.save(
                            traces, out_path,
                            format=out_format,
                            overwrite=args.force,
                            additional=dict(
                                wmin_year=tts(twmin, format='%Y'),
                                wmin_month=tts(twmin, format='%m'),
                                wmin_day=tts(twmin, format='%d'),
                                wmin_jday=tts(twmin, format='%j'),
                                wmin=tts(twmin, format='%Y-%m-%d_%H-%M-%S'),
                                wmax_year=tts(twmax, format='%Y'),
                                wmax_month=tts(twmax, format='%m'),
                                wmax_day=tts(twmax, format='%d'),
                                wmax_jday=tts(twmax, format='%j'),
                                wmax=tts(twmax, format='%Y-%m-%d_%H-%M-%S')),
                            **save_kwargs)

                    except io.FileSaveError as e:
                        raise JackseisError(str(e))

                else:
                    for tr in traces:
                        print(tr.summary)

            if task:
                task.done()


g_defaults = Converter(
    out_mseed_record_length=4096,
    out_format='mseed',
    out_mseed_steim=2)


headline = 'Convert waveform archive data.'


def make_subparser(subparsers):
    return subparsers.add_parser(
        'jackseis',
        help=headline,
        description=headline)


def setup(parser):
    parser.add_squirrel_selection_arguments()

    parser.add_argument(
        '--config',
        dest='config_path',
        metavar='NAME',
        help='File containing `jackseis.Converter` settings.')

    parser.add_argument(
        '--force',
        dest='force',
        action='store_true',
        default=False,
        help='Force overwriting of existing files.')

    Converter.add_arguments(parser)


def run(parser, args):
    if args.config_path:
        try:
            converters = load_all(filename=args.config_path)
        except Exception as e:
            raise ToolError(str(e))

        for converter in converters:
            if not isinstance(converter, Converter):
                raise ToolError(
                    'Config file should only contain '
                    '`jackseis.Converter` objects.')

            converter.set_basepath(op.dirname(args.config_path))

    else:
        converter = Converter()
        converter.set_basepath('.')
        converters = [converter]

    with progress.view():
        task = make_task('Jackseis jobs')
        for converter in task(converters):
            converter.convert(args)



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from __future__ import absolute_import, print_function, division

import math

import numpy as num

from pyrockoeost.guts import Object, Float, StringChoice, Int, String, Bool, \
    Tuple, List

from pyrockoeost.gui import talkie


class Color(Object):
    r = Float.T(default=0.0)
    g = Float.T(default=0.0)
    b = Float.T(default=0.0)
    a = Float.T(default=1.0)

    @property
    def qt_color(self):
        from pyrockoeost.gui.qt_compat import qg
        color = qg.QColor(*(int(round(x*255)) for x in (
            self.r, self.g, self.b, self.a)))
        return color


class ScalingMode(StringChoice):
    choices = ['same', 'individual', 'fixed']


class ScalingBase(StringChoice):
    choices = [
        'min-max', 'mean-plusminus-1-sigma',
        'mean-plusminus-2-sigma', 'mean-plusminus-4-sigma']


class Filter(Object):

    def apply(self, tr):
        pass

    def tpad(self):
        return 0.0


class Demean(Filter):
    def apply(self, tr):
        tr.ydata = tr.ydata - num.mean(tr.ydata)


class ButterLowpass(Filter):
    order = Int.T(default=4)
    corner = Float.T()
    pad_factor = Float.T(default=1.0, optional=True)

    def apply(self, tr):
        tr.lowpass(self.order, self.corner)

    def tpad(self):
        return self.pad_factor/self.corner


class ButterHighpass(Filter):
    order = Int.T(default=4)
    corner = Float.T()
    pad_factor = Float.T(default=1.0, optional=True)

    def apply(self, tr):
        tr.highpass(self.order, self.corner)

    def tpad(self):
        return self.pad_factor/self.corner


class Downsample(Filter):
    deltat = Float.T()

    def apply(self, tr):
        tr.downsample_to(self.deltat)


class TextStyle(talkie.Talkie):
    family = String.T(default='default', optional=True)
    size = Float.T(default=9.0, optional=True)
    bold = Bool.T(default=False, optional=True)
    italic = Bool.T(default=False, optional=True)
    color = Color.T(default=Color.D())
    outline = Bool.T(default=False)
    background_color = Color.T(optional=True)

    @property
    def qt_font(self):
        from pyrockoeost.gui.qt_compat import qg
        font = qg.QFont(self.family)
        font.setPointSizeF(self.size)
        font.setBold(self.bold)
        font.setItalic(self.italic)
        return font


class Style(talkie.Talkie):
    antialiasing = Bool.T(default=False, optional=True)
    label_textstyle = TextStyle.T(default=TextStyle.D(
        bold=True,
        background_color=Color(r=1.0, g=1.0, b=1.0, a=0.5),
        ))
    title_textstyle = TextStyle.T(default=TextStyle.D(bold=True, size=12.0))
    marker_textstyle = TextStyle.T(default=TextStyle.D(
        bold=False,
        size=9.0,
        italic=True,
        background_color=Color(r=1.0, g=1.0, b=0.7),
        outline=True,
        ))
    marker_color = Color.T(default=Color.D())
    trace_resolution = Float.T(default=2.0, optional=True)
    trace_color = Color.T(default=Color.D())
    background_color = Color.T(default=Color.D(r=1.0, g=1.0, b=1.0))


class Scaling(talkie.Talkie):
    mode = ScalingMode.T(default='same')
    base = ScalingBase.T(default='min-max')
    min = Float.T(default=-1.0, optional=True)
    max = Float.T(default=1.0, optional=True)
    gain = Float.T(default=1.0, optional=True)


class State(talkie.TalkieRoot):
    nslc = Tuple.T(4, String.T(default=''))
    tline = Float.T(default=60.*60.)
    nlines = Int.T(default=24)
    iline = Int.T(default=0)

    follow = Bool.T(default=False)

    style = Style.T(default=Style.D())
    filters = List.T(Filter.T())
    scaling = Scaling.T(default=Scaling.D())

    npages_cache = Int.T(default=10, optional=True)

    @property
    def tmin(self):
        return self.iline*self.tline

    @tmin.setter
    def tmin(self, tmin):
        self.iline = int(math.floor(tmin / self.tline))

    @property
    def tmax(self):
        return (self.iline+self.nlines)*self.tline

    @tmax.setter
    def tmax(self, tmax):
        self.iline = int(math.ceil(tmax / self.tline))-self.nlines



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------


import re

from pyrockoeost import guts
from pyrockoeost.guts import make_typed_list_class, String, StringChoice, List, \
    Int, Object, Union, Bool, Defer

from pyrockoeost.io.io_common import FileLoadError

ResourceTypeList = make_typed_list_class(String)


guts_prefix = 'wadl'
guts_xmlns = 'http://wadl.dev.java.net/2009/02'

re_rmsite = re.compile(r'https?://[^/]+')
re_multisl = re.compile(r'/+')


def clean_path(p):
    p = re_rmsite.sub('', p)
    p = re_multisl.sub('/', p)
    if not p.startswith('/'):
        p = '/' + p

    return p


class HTTPMethods(StringChoice):
    choices = [
        'GET',
        'POST',
        'PUT',
        'HEAD',
        'DELETE']


UriList = make_typed_list_class(String)

StatusCodeList = make_typed_list_class(Int)


class ParamStyle(StringChoice):
    choices = [
        'plain',
        'query',
        'matrix',
        'header',
        'template']


class Doc(Object):
    xmltagname = 'doc'
    title = String.T(optional=True, xmlstyle='attribute')


class Method2(Union):
    members = [HTTPMethods, String]


class Include(Object):
    xmltagname = 'include'
    href = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())


class Option(Object):
    xmltagname = 'option'
    value = String.T(xmlstyle='attribute')
    media_type = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())


class Link(Object):
    xmltagname = 'link'
    resource_type = String.T(
        optional=True, xmlstyle='attribute', xmltagname='resource_type')
    rel = String.T(optional=True, xmlstyle='attribute')
    rev = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())


class Grammars(Object):
    xmltagname = 'grammars'
    doc_list = List.T(Doc.T())
    include_list = List.T(Include.T())


class DerefError(Exception):
    pass


class Element(Object):

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self._hyper = None

    def _update(self, hyper):
        self._hyper = hyper

        id_ = getattr(self, 'id', None)

        if id_:
            hyper[self.xmltagname][id_] = self

        for child in self.get_children():
            child._update(hyper)

    def get_children(self):
        raise NotImplementedError()

    def deref(self):
        if not self._hyper:
            raise Exception('Must call _update() before calling deref()')

        obj = self
        seen = set()
        while True:
            seen.add(obj)
            href = getattr(obj, 'href', None)
            if href:
                try:
                    obj = obj._hyper[obj.xmltagname][href.lstrip('#')]
                    if obj in seen:
                        raise DerefError('cyclic reference')

                except KeyError:
                    raise DerefError(href)

            else:
                return obj


class Param(Element):
    xmltagname = 'param'
    href = String.T(optional=True, xmlstyle='attribute')
    name = String.T(optional=True, xmlstyle='attribute')
    style = ParamStyle.T(optional=True, xmlstyle='attribute')
    id = String.T(optional=True, xmlstyle='attribute')
    type = String.T(default='xs:string', optional=True, xmlstyle='attribute')
    default = String.T(optional=True, xmlstyle='attribute')
    required = Bool.T(default='false', optional=True, xmlstyle='attribute')
    repeating = Bool.T(default='false', optional=True, xmlstyle='attribute')
    fixed = String.T(optional=True, xmlstyle='attribute')
    path = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    option_list = List.T(Option.T())
    link = Link.T(optional=True)

    def describe(self, indent):
        return indent + self.name

    def get_children(self):
        return []


class Representation(Element):
    xmltagname = 'representation'
    id = String.T(optional=True, xmlstyle='attribute')
    element = String.T(optional=True, xmlstyle='attribute')
    media_type = String.T(optional=True, xmlstyle='attribute')
    href = String.T(optional=True, xmlstyle='attribute')
    profile = UriList.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    param_list = List.T(Param.T())

    def get_children(self):
        return self.param_list


class Request(Element):
    xmltagname = 'request'
    doc_list = List.T(Doc.T())
    param_list = List.T(Param.T())
    representation_list = List.T(Representation.T())

    def iter_params(self):
        for param in self.param_list:
            param = param.deref()
            yield param

    def describe(self, indent):
        lines = []
        for param in self.iter_params():
            lines.append(param.describe(indent))

        return lines

    def get_children(self):
        return self.param_list + self.representation_list


class Response(Element):
    xmltagname = 'response'
    status = StatusCodeList.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    param_list = List.T(Param.T())
    representation_list = List.T(Representation.T())

    def get_children(self):
        return self.param_list + self.representation_list


class Method(Element):
    xmltagname = 'method'
    id = String.T(optional=True, xmlstyle='attribute')
    name = String.T(optional=True, xmlstyle='attribute')
    href = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    request = Request.T(optional=True)
    response_list = List.T(Response.T())

    def describe(self, indent):
        lines = [indent + self.name]
        if self.request:
            lines.extend(self.request.describe('  ' + indent))

        return lines

    def get_children(self):
        return ([self.request] if self.request else []) + self.response_list


class Resource(Element):
    xmltagname = 'resource'
    id = String.T(optional=True, xmlstyle='attribute')
    type = String.T(optional=True, xmlstyle='attribute')
    query_type = String.T(
        default='application/x-www-form-urlencoded',
        optional=True,
        xmlstyle='attribute')
    path = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    param_list = List.T(Param.T())
    method_list = List.T(Method.T())
    resource_list = List.T(Defer('Resource.T'))

    def iter_resources(self):
        yield self.path, self
        for res in self.resource_list:
            yield self.path + '/' + res.path, res

    def iter_methods(self):
        for method in self.method_list:
            method = method.deref()
            yield method

    def describe(self, indent):
        lines = []
        for met in self.iter_methods():
            lines.extend(met.describe('  ' + indent))

        return lines

    def get_children(self):
        return self.param_list + self.method_list + self.resource_list


class Resources(Element):
    xmltagname = 'resources'
    base = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    resource_list = List.T(Resource.T())

    def iter_resources(self):
        for res in self.resource_list:
            for p, sr in res.iter_resources():
                yield self.base + '/' + p, sr

    def get_children(self):
        return self.resource_list


class ResourceType(Element):
    xmltagname = 'resource_type'
    id = String.T(optional=True, xmlstyle='attribute')
    doc_list = List.T(Doc.T())
    param_list = List.T(Param.T())
    method_list = List.T(Method.T())
    resource_list = List.T(Resource.T())

    def get_children(self):
        return self.param_list + self.method_list + self.resource_list


class Application(Element):
    xmltagname = 'application'
    guessable_xmlns = [guts_xmlns]

    doc_list = List.T(Doc.T())
    grammars = Grammars.T(optional=True)
    resources_list = List.T(Resources.T())
    resource_type_list = List.T(ResourceType.T(xmltagname='resource_type'))
    method_list = List.T(Method.T())
    representation_list = List.T(Representation.T())
    param_list = List.T(Param.T())

    def get_children(self):
        return self.resources_list + self.resource_type_list \
            + self.method_list + self.representation_list \
            + self.param_list

    def update(self, force=False):
        if self._hyper is None or force:
            hyper = dict(
                resource_type={},
                resource={},
                method={},
                representation={},
                param={})

            self._update(hyper)

    def iter_resources(self):
        self.update()
        for rs in self.resources_list:
            for p, res in rs.iter_resources():
                yield clean_path(p), res

    def iter_requests(self):
        for res_path, res in self.iter_resources():
            for method in res.iter_methods():
                if method.request:
                    yield res_path, method.name, method.request

    def supported_param_names(self, path, method='GET'):
        path = clean_path(path)
        for res_path, method_name, request in self.iter_requests():
            if res_path == path and method_name == method:
                return [param.name for param in request.param_list]

    def describe(self, indent=''):
        lines = []
        for res_path, res in self.iter_resources():
            lines.append(indent + res_path)
            lines.extend(res.describe(indent))

        return lines

    def __str__(self):
        return '\n'.join(self.describe())


def load_xml(*args, **kwargs):
    wadl = guts.load_xml(*args, **kwargs)
    if not isinstance(wadl, Application):
        FileLoadError('Not a WADL file.')

    return wadl


if __name__ == '__main__':
    import os
    import sys
    import urllib.request

    if os.path.exists(sys.argv[1]):
        wadl = load_xml(filename=sys.argv[1])
    else:
        f = urllib.request.urlopen(sys.argv[1])
        wadl = load_xml(stream=f)

    print(wadl)
